import logging
from pathlib import Path
from typing import Optional

from riot_na.alignment.alignment_utils import offset_alignments
from riot_na.alignment.gene_aligner import (
    GeneAligner,
    create_aligner,
    create_v_gene_aligner,
)
from riot_na.alignment.segment_gene_aligner import SegmentGeneAligner
from riot_na.alignment.segment_gene_aligner import (
    create_v_gene_aligner as create_segment_v_gene_aligner,
)
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AlignmentSegment,
    AlignmentsNT,
    GermlineGene,
    Locus,
    Organism,
)

logger = logging.getLogger(__name__)


class VDJCAlignerNT:
    def __init__(
        self,
        v_aligner: SegmentGeneAligner | GeneAligner,
        d_aligners: dict[Organism, GeneAligner],
        j_aligners: dict[Organism, dict[Locus, GeneAligner]],
        c_aligners: dict[Organism, dict[Locus, GeneAligner]],
    ) -> None:
        self.v_aligner = v_aligner
        self.d_aligners = d_aligners
        self.j_aligners = j_aligners
        self.c_aligners = c_aligners  # at the time of writing, there is only one c germline database

    def produce_nt_alignments(self, query: str) -> list[AlignmentsNT]:
        results = []

        v_alignments = self.v_aligner.align(query, both_strains=True)
        for v_alignment in v_alignments:
            try:
                aligments_nt = self._produce_vdjc_nt_alignment(v_alignment)
                results.append(aligments_nt)
            except ValueError as e:
                logger.error(
                    "Error aligning %s: %s",
                    v_alignment.best_alignment.q_seq if v_alignment.best_alignment else "unknown sequence",
                    e,
                )
        return results

    def _produce_vdjc_nt_alignment(self, v_alignment_segment: AlignmentSegment) -> AlignmentsNT:
        aligments_nt = AlignmentsNT()
        aligments_nt.segment_start = v_alignment_segment.segment_start
        aligments_nt.segment_end = v_alignment_segment.segment_end

        v_alignment = v_alignment_segment.best_alignment
        if v_alignment is None:
            return aligments_nt

        query_v = v_alignment.q_seq
        aligments_nt.v = v_alignment

        # mask alignments
        query_j = v_alignment.q_seq[v_alignment.q_end :]

        # align j genes
        species = v_alignment.species
        locus = v_alignment.locus

        j_aligner = self.j_aligners[species][locus]
        j_alignment = j_aligner.align(query_j, both_strains=False)[0].best_alignment

        if j_alignment is None:
            return aligments_nt

        j_alignment = offset_alignments(v_alignment.q_end, j_alignment)
        assert j_alignment is not None
        aligments_nt.j = j_alignment

        # mask c input (query_db_clean -> valn_end + jaln_end)
        query_c = query_v[j_alignment.q_end :]

        if query_c:
            # align c
            c_aligner = self.c_aligners[species][locus]
            c_alignment = c_aligner.align(query_c, both_strains=False)[0].best_alignment

            if c_alignment is not None:
                c_alignment = offset_alignments(j_alignment.q_end, c_alignment)
                aligments_nt.c = c_alignment

        if locus == Locus.IGH:
            # mask d input (query_db_clean -> valn_end + jaln_end)
            query_d = query_v[v_alignment.q_end : j_alignment.q_start]
            if query_d:
                # align d genes
                d_aligner = self.d_aligners[species]
                d_alignment = d_aligner.align(query_d, both_strains=False)[0].best_alignment

                if d_alignment is not None:
                    d_alignment = offset_alignments(v_alignment.q_end, d_alignment)
                    aligments_nt.d = d_alignment

        return aligments_nt


def create_vdjc_aligner_nt(
    allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR, use_segment_aligner: bool = False
):
    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS, Organism.VICUGNA_PACOS]

    v_aligner = (
        create_segment_v_gene_aligner(allowed_species=allowed_species, db_dir=db_dir)
        if use_segment_aligner
        else create_v_gene_aligner(allowed_species=allowed_species, db_dir=db_dir)
    )

    d_aligners = {}
    for organism in allowed_species:
        d_aligner = create_aligner(organism=organism, germline_gene=GermlineGene.D, locus=Locus.IGH, db_dir=db_dir)
        d_aligners[organism] = d_aligner

    j_aligners: dict[Organism, dict[Locus, GeneAligner]] = {}

    for organism in allowed_species:
        organism_aligners = j_aligners.get(organism, {})

        for locus in Locus:
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            j_aligner = create_aligner(organism=organism, germline_gene=GermlineGene.J, locus=locus, db_dir=db_dir)
            organism_aligners[locus] = j_aligner

        j_aligners[organism] = organism_aligners

    c_aligners: dict[Organism, dict[Locus, GeneAligner]] = {}

    for organism in allowed_species:

        organism_aligners = c_aligners.get(organism, {})

        if organism not in [Organism.HOMO_SAPIENS, Organism.VICUGNA_PACOS, Organism.CUSTOM]:
            c_organism = Organism.HOMO_SAPIENS
        else:
            c_organism = organism

        for locus in Locus:
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            c_aligner = create_aligner(organism=c_organism, germline_gene=GermlineGene.C, locus=locus, db_dir=db_dir)
            organism_aligners[locus] = c_aligner

        c_aligners[organism] = organism_aligners

    vdjc_aligner_nt = VDJCAlignerNT(v_aligner, d_aligners, j_aligners, c_aligners)

    return vdjc_aligner_nt



# pylint: skip-file
from riot_na.alignment.nt_gene_alignments import create_vdjc_aligner_nt
from riot_na.data.model import Locus, Organism


def test_produce_nt_alignments():

    # given
    aligner = create_vdjc_aligner_nt(use_segment_aligner=True)

    query_sequence = "AACAACACATGTCCAATGTCCTCTCCACAGACACTGAACACACTGACTCTAACCATGGGAAGGAGCTGGATCTTTCTCTTCCTCCTGTCAGGAACTGCAGGTGTCCACTCTGAGGTCCAGCTGCAACAGTCTGGACCTGTGCTGGTGAAGCCTGGGGCTTCAGTGAAGATGTCCTGTAAGGCTTCTGGATACACATTCACTGACTACTATATGAACTGGGTGAAGCAGAGCCATGGAAAGAGACTTGAGTGGATTGGAGTTATTAATCCTTACAACGGTGGTACTAACTATAACCAGAAGTTCAAGGGCAAGGCCACATTGACTGTTGACAAGTCCTCCAGCACAGCCTACATGGAGCTCAACAGCCTGACATCTGAGGACTCTGCAGTCTATTACTGTGCAGATGGGATTATTACGAATTGGTATTTCGATGTCTGGGGCACAGGGACCACGGTCACCGTCTCCTCAGCCAAAACGACACCCCCATCTGTCTATCCACTGGCCCCTGGATCTGCTGCCCAAACTAACTCCATGGTGACCCTGGGATGCCTGGTCAAGGGCTATTTCCCTGAGCCAGTGACAGTGACCTGGAACTCTGGATCCCTGTCCAGCGGTGTGCACACCTTCCCAGCTGTCCTGCAGTCTGACCTCTACACTCTGAGCAGCTCAGTGACTGTCCCCTCCAGCACCTGGCCCAGCCAGACCGTCACCTGCAACGTTGCCCACCCGGCCAGCAGCACCAAGGTGGACAAGAAAATTGTGCCCAGGGATTGTGGTTGTAAGCCTTGCATATGTACAGTCCCAGAAGT"

    # when
    alignments = aligner.produce_nt_alignments(query_sequence)[0]

    # then
    assert alignments.v is not None
    assert alignments.v.target_id == "IGHV-2ZSK"
    assert alignments.v.alignment_score == 445.8980348533401
    assert alignments.v.seq_identity == 0.9896907216494846
    assert alignments.v.e_value == 4.962993038125875e-78
    assert alignments.segment_start == 111
    assert alignments.v.q_start == 0
    assert alignments.v.q_end == 402 - alignments.segment_start
    assert alignments.v.q_len == 291
    assert alignments.v.t_start == 0
    assert alignments.v.t_end == 291
    assert alignments.v.t_len == 291
    assert alignments.v.cigar == "291M"
    assert alignments.v.rev_comp is False
    assert alignments.v.species == Organism.MUS_MUSCULUS
    assert alignments.v.locus == Locus.IGH
    assert (
        alignments.v.q_seq
        == "AACAACACATGTCCAATGTCCTCTCCACAGACACTGAACACACTGACTCTAACCATGGGAAGGAGCTGGATCTTTCTCTTCCTCCTGTCAGGAACTGCAGGTGTCCACTCTGAGGTCCAGCTGCAACAGTCTGGACCTGTGCTGGTGAAGCCTGGGGCTTCAGTGAAGATGTCCTGTAAGGCTTCTGGATACACATTCACTGACTACTATATGAACTGGGTGAAGCAGAGCCATGGAAAGAGACTTGAGTGGATTGGAGTTATTAATCCTTACAACGGTGGTACTAACTATAACCAGAAGTTCAAGGGCAAGGCCACATTGACTGTTGACAAGTCCTCCAGCACAGCCTACATGGAGCTCAACAGCCTGACATCTGAGGACTCTGCAGTCTATTACTGTGCAGATGGGATTATTACGAATTGGTATTTCGATGTCTGGGGCACAGGGACCACGGTCACCGTCTCCTCAGCCAAAACGACACCCCCATCTGTCTATCCACTGGCCCCTGGATCTGCTGCCCAAACTAACTCCATGGTGACCCTGGGATGCCTGGTCAAGGGCTATTTCCCTGAGCCAGTGACAGTGACCTGGAACTCTGGATCCCTGTCCAGCGGTGTGCACACCTTCCCAGCTGTCCTGCAGTCTGACCTCTACACTCTGAGCAGCTCAGTGACTGTCCCCTCCAGCACCTGGCCCAGCCAGACCGTCACCTGCAACGTTGCCCACCCGGCCAGCAGCACCAAGGTGGACAAGAAAATTGTGCCCAGGGATTGTGGTTGTAAGCCTTGCATATGTACAGTCCCAGAAGT"[
            alignments.segment_start : alignments.segment_end
        ]
    )
    assert (
        alignments.v.t_seq
        == "GAGGTCCAGCTGCAACAGTCTGGACCTGTGCTGGTGAAGCCTGGGGCTTCAGTGAAGATGTCCTGTAAGGCTTCTGGATACACATTCACTGACTACTATATGAACTGGGTGAAGCAGAGCCATGGAAAGAGCCTTGAGTGGATTGGAGTTATTAATCCTTACAACGGTGGTACTAGCTACAACCAGAAGTTCAAGGGCAAGGCCACATTGACTGTTGACAAGTCCTCCAGCACAGCCTACATGGAGCTCAACAGCCTGACATCTGAGGACTCTGCAGTCTATTACTGTGCAAG"
    )
    assert alignments.v.reading_frame == 0

    assert alignments.d is not None
    assert alignments.d.target_id == "IGHD-4RNO"
    assert alignments.d.alignment_score == 14.301386420997765
    assert alignments.d.seq_identity == 0.9
    assert alignments.d.e_value == 24.1875
    assert alignments.d.q_start == 408 - alignments.segment_start
    assert alignments.d.q_end == 418 - alignments.segment_start
    assert alignments.d.q_len == 10
    assert alignments.d.t_start == 6
    assert alignments.d.t_end == 16
    assert alignments.d.t_len == 10
    assert alignments.d.cigar == "10M"
    assert alignments.d.rev_comp is False
    assert alignments.d.species == Organism.MUS_MUSCULUS
    assert alignments.d.locus == Locus.IGH
    assert alignments.d.q_seq == "GATGGGATTATTACGAAT"
    assert alignments.d.t_seq == "TCTACTATGATTACGAC"

    assert alignments.j is not None
    assert alignments.j.target_id == "IGHJ-32C2"
    assert alignments.j.alignment_score == 75.06770154324091
    assert alignments.j.seq_identity == 0.9795918367346939
    assert alignments.j.e_value == 8.791403161012568e-10
    assert alignments.j.q_start == 420 - alignments.segment_start
    assert alignments.j.q_end == 469 - alignments.segment_start
    assert alignments.j.q_len == 49
    assert alignments.j.t_start == 4
    assert alignments.j.t_end == 53
    assert alignments.j.t_len == 49
    assert alignments.j.cigar == "49M"
    assert alignments.j.rev_comp is False
    assert alignments.j.species == Organism.MUS_MUSCULUS
    assert alignments.j.locus == Locus.IGH
    assert (
        alignments.j.q_seq
        == "GATGGGATTATTACGAATTGGTATTTCGATGTCTGGGGCACAGGGACCACGGTCACCGTCTCCTCAGCCAAAACGACACCCCCATCTGTCTATCCACTGGCCCCTGGATCTGCTGCCCAAACTAACTCCATGGTGACCCTGGGATGCCTGGTCAAGGGCTATTTCCCTGAGCCAGTGACAGTGACCTGGAACTCTGGATCCCTGTCCAGCGGTGTGCACACCTTCCCAGCTGTCCTGCAGTCTGACCTCTACACTCTGAGCAGCTCAGTGACTGTCCCCTCCAGCACCTGGCCCAGCCAGACCGTCACCTGCAACGTTGCCCACCCGGCCAGCAGCACCAAGGTGGACAAGAAAATTGTGCCCAGGGATTGTGGTTGTAAGCCTTGCATATGTACAGTCCCAGAAGT"
    )
    assert alignments.j.t_seq == "CTACTGGTACTTCGATGTCTGGGGCACAGGGACCACGGTCACCGTCTCCTCAG"
    assert alignments.j.reading_frame == 1

    # AlignmentEntryNT(target_id='IGHG1', alignment_score=209.06521694100786, seq_identity=0.7833333333333333, e_value=3.1801713964550777e-34, q_start=538, q_end=775, q_len=237, t_start=70, t_end=310, t_len=240, cigar='104M3D133M', rev_comp=False, species=<Organism.HOMO_SAPIENS: 'human'>, locus=<Locus.IGH: 'igh'>, q_seq='CCAAAACGACACCCCCATCTGTCTATCCACTGGCCCCTGGATCTGCTGCCCAAACTAACTCCATGGTGACCCTGGGATGCCTGGTCAAGGGCTATTTCCCTGAGCCAGTGACAGTGACCTGGAACTCTGGATCCCTGTCCAGCGGTGTGCACACCTTCCCAGCTGTCCTGCAGTCTGACCTCTACACTCTGAGCAGCTCAGTGACTGTCCCCTCCAGCACCTGGCCCAGCCAGACCGTCACCTGCAACGTTGCCCACCCGGCCAGCAGCACCAAGGTGGACAAGAAAATTGTGCCCAGGGATTGTGGTTGTAAGCCTTGCATATGTACAGTCCCAGAAGT', t_seq='GCCTCCACCAAGGGCCCATCGGTCTTCCCCCTGGCACCCTCCTCCAAGAGCACCTCTGGGGGCACAGCAGCCCTGGGCTGCCTGGTCAAGGACTACTTCCCCGAACCGGTGACGGTGTCGTGGAACTCAGGCGCCCTGACCAGCGGCGTGCACACCTTCCCGGCTGTCCTACAGTCCTCAGGACTCTACTCCCTCAGCAGCGTGGTGACCGTGCCCTCCAGCAGCTTGGGCACCCAGACCTACATCTGCAACGTGAATCACAAGCCCAGCAACACCAAGGTGGACAAGAAAGTTGAGCCCAAATCTTGTGACAAAACTCACACATGCCCACCGTGCCCAGCACCTGAACTCCTGGGGGGACCGTCAGTCTTCCTCTTCCCCCCAAAACCCAAGGACACCCTCATGATCTCCCGGACCCCTGAGGTCACATGCGTGGTGGTGGACGTGAGCCACGAAGACCCTGAGGTCAAGTTCAACTGGTACGTGGACGGCGTGGAGGTGCATAATGCCAAGACAAAGCCGCGGGAGGAGCAGTACAACAGCACGTACCGTGTGGTCAGCGTCCTCACCGTCCTGCACCAGGACTGGCTGAATGGCAAGGAGTACAAGTGCAAGGTCTCCAACAAAGCCCTCCCAGCCCCCATCGAGAAAACCATCTCCAAAGCCAAAGGGCAGCCCCGAGAACCACAGGTGTACACCCTGCCCCCATCCCGGGATGAGCTGACCAAGAACCAGGTCAGCCTGACCTGCCTGGTCAAAGGCTTCTATCCCAGCGACATCGCCGTGGAGTGGGAGAGCAATGGGCAGCCGGAGAACAACTACAAGACCACGCCTCCCGTGCTGGACTCCGACGGCTCCTTCTTCCTCTACAGCAAGCTCACCGTGGACAAGAGCAGGTGGCAGCAGGGGAACGTCTTCTCATGCTCCGTGATGCATGAGGCTCTGCACAACCACTACACACAGAAGAGCCTCTCCCTGTCTCCGGGTAAATGA', reading_frame=None))
    assert alignments.c is not None
    assert alignments.c.target_id == "IGHG1"
    assert alignments.c.alignment_score == 209.06521694100786
    assert alignments.c.seq_identity == 0.7833333333333333
    assert alignments.c.e_value == 3.1801713964550777e-34
    assert alignments.c.q_start == 538 - alignments.segment_start
    assert alignments.c.q_end == 775 - alignments.segment_start
    assert alignments.c.q_len == 237
    assert alignments.c.t_start == 70
    assert alignments.c.t_end == 310
    assert alignments.c.t_len == 240
    assert alignments.c.cigar == "104M3D133M"
    assert alignments.c.rev_comp is False
    assert alignments.c.species == Organism.HOMO_SAPIENS
    assert alignments.c.locus == Locus.IGH
    assert (
        alignments.c.q_seq
        == "CCAAAACGACACCCCCATCTGTCTATCCACTGGCCCCTGGATCTGCTGCCCAAACTAACTCCATGGTGACCCTGGGATGCCTGGTCAAGGGCTATTTCCCTGAGCCAGTGACAGTGACCTGGAACTCTGGATCCCTGTCCAGCGGTGTGCACACCTTCCCAGCTGTCCTGCAGTCTGACCTCTACACTCTGAGCAGCTCAGTGACTGTCCCCTCCAGCACCTGGCCCAGCCAGACCGTCACCTGCAACGTTGCCCACCCGGCCAGCAGCACCAAGGTGGACAAGAAAATTGTGCCCAGGGATTGTGGTTGTAAGCCTTGCATATGTACAGTCCCAGAAGT"
    )
    assert (
        alignments.c.t_seq
        == "GCCTCCACCAAGGGCCCATCGGTCTTCCCCCTGGCACCCTCCTCCAAGAGCACCTCTGGGGGCACAGCAGCCCTGGGCTGCCTGGTCAAGGACTACTTCCCCGAACCGGTGACGGTGTCGTGGAACTCAGGCGCCCTGACCAGCGGCGTGCACACCTTCCCGGCTGTCCTACAGTCCTCAGGACTCTACTCCCTCAGCAGCGTGGTGACCGTGCCCTCCAGCAGCTTGGGCACCCAGACCTACATCTGCAACGTGAATCACAAGCCCAGCAACACCAAGGTGGACAAGAAAGTTGAGCCCAAATCTTGTGACAAAACTCACACATGCCCACCGTGCCCAGCACCTGAACTCCTGGGGGGACCGTCAGTCTTCCTCTTCCCCCCAAAACCCAAGGACACCCTCATGATCTCCCGGACCCCTGAGGTCACATGCGTGGTGGTGGACGTGAGCCACGAAGACCCTGAGGTCAAGTTCAACTGGTACGTGGACGGCGTGGAGGTGCATAATGCCAAGACAAAGCCGCGGGAGGAGCAGTACAACAGCACGTACCGTGTGGTCAGCGTCCTCACCGTCCTGCACCAGGACTGGCTGAATGGCAAGGAGTACAAGTGCAAGGTCTCCAACAAAGCCCTCCCAGCCCCCATCGAGAAAACCATCTCCAAAGCCAAAGGGCAGCCCCGAGAACCACAGGTGTACACCCTGCCCCCATCCCGGGATGAGCTGACCAAGAACCAGGTCAGCCTGACCTGCCTGGTCAAAGGCTTCTATCCCAGCGACATCGCCGTGGAGTGGGAGAGCAATGGGCAGCCGGAGAACAACTACAAGACCACGCCTCCCGTGCTGGACTCCGACGGCTCCTTCTTCCTCTACAGCAAGCTCACCGTGGACAAGAGCAGGTGGCAGCAGGGGAACGTCTTCTCATGCTCCGTGATGCATGAGGCTCTGCACAACCACTACACACAGAAGAGCCTCTCCCTGTCTCCGGGTAAATGA"
    )
    assert alignments.c.reading_frame is None


if __name__ == "__main__":
    test_produce_nt_alignments()



from riot_na.api.utils import map_insertion_number_to_letter


def test_map_insertion_number_to_letter_no_decimal():
    # given
    position = "111"

    # when
    result = map_insertion_number_to_letter(position)

    # then
    assert result == "111"


def test_map_insertion_number_to_letter_decimal():
    # given
    position = "111.1"

    # when
    result = map_insertion_number_to_letter(position)

    # then
    assert result == "111A"


def test_map_insertion_number_to_letter_z():
    # given
    position = "111.26"

    # when
    result = map_insertion_number_to_letter(position)

    # then
    assert result == "111Z"


def test_map_insertion_number_to_letter_double_letter():
    # given
    position = "111.27"

    # when
    result = map_insertion_number_to_letter(position)

    # then
    assert result == "111AA"


def test_map_insertion_number_to_letter_large_insertion():
    # given
    position = "111.100"

    # when
    result = map_insertion_number_to_letter(position)

    # then
    assert result == "111CV"



from typing import Sequence

from riot_na import Prefiltering
from riot_na.data.model import (
    Gene,
    GeneAA,
    Locus,
    Organism,
    SegmentMatch,
    SpeciesGeneMatch,
    SpeciesPrefilteringSegmentResult,
)
from riot_na.riot_na import (  # pylint: disable=no-name-in-module
    PrefilteringSegmentResult,
)


class MultiSpeciesSegmentPrefiltering:
    def __init__(
        self,
        all_genes: Sequence[Gene | GeneAA],
        kmer_size: int,
        distance_threshold: int,
        top_n: int,
        modulo_n: int,
        min_segment_length: int = 30,
        min_coverage: int = 20,
    ):
        self.all_genes = dict(
            map(lambda gene: (gene.species + "|" + gene.locus + "|" + gene.name, gene.sequence), all_genes)
        )

        self.prefiltering = Prefiltering(
            self.all_genes,
            top_n=top_n,
            kmer_size=kmer_size,
            distance_threshold=distance_threshold,
            modulo_n=modulo_n,
            min_segment_length=min_segment_length,
            min_coverage=min_coverage,
        )

    def calculate_segment_matches_with_rev_comp(self, query: str) -> SpeciesPrefilteringSegmentResult:
        try:
            raw_result = self.prefiltering.calculate_segment_matches_with_rev_comp(query)
        except ValueError:
            return SpeciesPrefilteringSegmentResult(query=query, rev_comp_query="", segments=[])

        return self._map_raw_result_to_segments(raw_result)

    def calculate_segment_matches(self, query: str) -> SpeciesPrefilteringSegmentResult:
        try:
            raw_result = self.prefiltering.calculate_segment_matches(query)
        except ValueError:
            return SpeciesPrefilteringSegmentResult(query=query, rev_comp_query="", segments=[])

        return self._map_raw_result_to_segments(raw_result)

    def _map_raw_result_to_segments(self, raw_result: PrefilteringSegmentResult) -> SpeciesPrefilteringSegmentResult:
        segments = []
        for segment in raw_result.segments:
            mapped_matches = []
            for match in segment.matching_genes:
                (species, locus, name) = match.gene_id.split("|")
                mapped_matches.append(
                    SpeciesGeneMatch(
                        species_gene_id=match.gene_id,
                        gene_id=name,
                        rev_comp=match.rev_comp,
                        coverage=match.coverage,
                        species=Organism(species),
                        locus=Locus(locus),
                    )
                )
            segments.append(
                SegmentMatch(
                    segment_start=segment.segment_start,
                    segment_length=segment.query_length(),
                    query_start=segment.query_start,
                    query_end=segment.query_end,
                    coverage=segment.coverage,
                    match_count=segment.match_count,
                    matching_genes=mapped_matches,
                )
            )
        return SpeciesPrefilteringSegmentResult(
            query=raw_result.query, rev_comp_query=raw_result.rev_comp_query, segments=segments
        )



from riot_na.alignment.aa_gene_alignments import create_vjc_aligner_aa
from riot_na.data.model import Locus, Organism


def test_produce_aa_alignments():
    # given
    aligner = create_vjc_aligner_aa(use_segment_aligner=True)

    query_sequence = "QVQLQQWGAGLLKPSETLSLTCAVFGGSFSGYYWSWIRQPPGKGLEWIGEINHRGNTNDNPSLKSRVTISVDTSKNQFALKLSSVTAADTAVYYCARERGYTYGNFDHWGQGTLVTVSSASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPGK"

    # when
    alignments = aligner.produce_aa_alignments(query_sequence)[0]

    # then
    assert alignments.v is not None
    assert alignments.v.target_id == "IGHV4-34*01"
    assert alignments.v.alignment_score == 192.68417758360758
    assert alignments.v.seq_identity == 0.9484536082474226
    assert alignments.v.e_value == 2.2632803302266517e-141
    assert alignments.v.q_start == 0
    assert alignments.v.q_end == 97
    assert alignments.v.t_start == 0
    assert alignments.v.t_end == 97
    assert alignments.v.cigar == "97M"
    assert alignments.v.species == Organism.HOMO_SAPIENS
    assert alignments.v.locus == Locus.IGH
    assert (
        alignments.v.q_seq
        == "QVQLQQWGAGLLKPSETLSLTCAVFGGSFSGYYWSWIRQPPGKGLEWIGEINHRGNTNDNPSLKSRVTISVDTSKNQFALKLSSVTAADTAVYYCARERGYTYGNFDHWGQGTLVTVSSASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPGK"
    )
    assert (
        alignments.v.t_seq
        == "QVQLQQWGAGLLKPSETLSLTCAVYGGSFSGYYWSWIRQPPGKGLEWIGEINHSGSTNYNPSLKSRVTISVDTSKNQFSLKLSSVTAADTAVYYCAR"
    )

    assert alignments.j is not None
    assert alignments.j.target_id == "IGHJ1*01"
    assert alignments.j.alignment_score == 32.01169295201648
    assert alignments.j.seq_identity == 0.9285714285714286
    assert alignments.j.e_value == 9.01920682236379e-18
    assert alignments.j.q_start == 8
    assert alignments.j.q_end == 22
    assert alignments.j.t_start == 3
    assert alignments.j.t_end == 17
    assert alignments.j.cigar == "14M"
    assert alignments.j.species == Organism.HOMO_SAPIENS
    assert alignments.j.locus == Locus.IGH
    assert (
        alignments.j.q_seq
        == "ERGYTYGNFDHWGQGTLVTVSSASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPGK"
    )
    assert alignments.j.t_seq == "AEYFQHWGQGTLVTVSS"

    assert alignments.c is not None
    assert alignments.c.target_id == "IGHG1"
    assert alignments.c.alignment_score == 674.7016314783808
    assert alignments.c.seq_identity == 1.0
    assert alignments.c.e_value == 0.0
    assert alignments.c.q_start == 0
    assert alignments.c.q_end == 328
    assert alignments.c.t_start == 0
    assert alignments.c.t_end == 328
    assert alignments.c.cigar == "328M"
    assert alignments.c.species == Organism.HOMO_SAPIENS
    assert alignments.c.locus == Locus.IGH
    assert (
        alignments.c.q_seq
        == "ASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPGK"
    )
    assert (
        alignments.c.t_seq
        == "ASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPELQLEESCAEAQDGELDGLWTTITIFITLFLLSVCYSATVTFFKVKWIFSSVVDLKQTIIPDYRNMIGQGA"
    )


def test_produce_aa_alignments_therapeutic():
    # given
    aligner = create_vjc_aligner_aa(use_segment_aligner=True)

    query_sequence = "EVQLVESGGGSVQPGGSLRLSCTASGFTISRSYWICWVRQAPGKGLEWVGCIYGDNDITPLYANWAKGRFTISRDTSKNTVYLQMNSLRAEDTATYYCARLGYADYAYDLWGQGTTVTVSS"

    # when
    alignments = aligner.produce_aa_alignments(query_sequence)[0]

    # then
    assert alignments.v is not None
    assert alignments.v.target_id == "IGHV3-66*01"


if __name__ == "__main__":
    test_produce_aa_alignments()



import os
from functools import partial
from pathlib import Path
from typing import Optional, Sequence

from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.gene_aligner import AA_ALIGNER_PARAMS, ALIGNER_PARAMS
from riot_na.alignment.gene_aligner import (
    get_aa_aligner_params as gene_aa_aligner_params,
)
from riot_na.alignment.gene_aligner import get_aligner_params as gene_aligner_params
from riot_na.alignment.gene_aligner import (
    get_gene_parsing_function,
    parse_gene_aa,
    read_genes,
)
from riot_na.alignment.skbio_alignment import align, align_aa
from riot_na.common.gene_match_utils import create_gene_lookup
from riot_na.common.multi_species_segment_prefiltering import (
    MultiSpeciesSegmentPrefiltering,
)
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AlignmentEntryAA,
    AlignmentEntryNT,
    AlignmentSegment,
    Gene,
    GeneAA,
    GermlineGene,
    Locus,
    Organism,
    SegmentMatch,
    SpeciesPrefilteringSegmentResult,
)


class SegmentGeneAligner:
    """
    Segment-based gene aligner that provides alignments for each detected domain/segment.

    This aligner uses the segment-centric prefiltering approach to first identify
    non-overlapping segments in the query sequence, then performs detailed alignments
    for each segment against its best matching genes.

    Compatible with GeneAligner interface for easy replacement.
    """

    def __init__(
        self,
        genes: Sequence[Gene],
        kmer_size: int = 6,
        distance_threshold: int = 10,
        top_n: int = 10,
        modulo_n: int = 1,
        e_value_threshold: Optional[float] = None,
        alignment_length_threshold: Optional[int] = None,
        min_prefiltering_coverage: int = 30,
        min_segment_length: int = 30,
    ) -> None:
        # Initialize the Rust-based prefiltering
        self.prefiltering = MultiSpeciesSegmentPrefiltering(
            all_genes=genes,
            kmer_size=kmer_size,
            distance_threshold=distance_threshold,
            top_n=top_n,
            modulo_n=modulo_n,
            min_segment_length=min_segment_length,
            min_coverage=min_prefiltering_coverage,
        )

        # Store gene sequences for alignment
        self.gene_lookup = create_gene_lookup(genes)
        self.db_length = sum(len(gene.sequence) for gene in genes)

        # Alignment parameters
        self.e_value_threshold = e_value_threshold
        self.alignment_length_threshold = alignment_length_threshold

    # Compatibility method to match GeneAligner interface
    def _prefilter(self, query: str, both_strains: bool) -> SpeciesPrefilteringSegmentResult:
        """Prefilter method compatible with GeneAligner interface"""
        if both_strains:
            return self.prefiltering.calculate_segment_matches_with_rev_comp(query)
        return self.prefiltering.calculate_segment_matches(query)

    def _align_segment_genes(
        self, segment: SegmentMatch, query: str, rev_comp_query: str
    ) -> Sequence[AlignmentEntryNT]:
        """Align the top genes for a specific segment"""
        alignments = []

        for gene_info in segment.matching_genes:
            # Get the appropriate query sequence based on strand
            current_query = rev_comp_query if gene_info.rev_comp else query

            # Create aligner for this query
            aligner = StripedSmithWaterman(current_query, **ALIGNER_PARAMS)

            # Get target sequence
            target_gene = self.gene_lookup[gene_info.species_gene_id]

            # Perform alignment
            alignment = align(
                aligner=aligner,
                target_id=gene_info.gene_id,
                target=target_gene.sequence,
                query=current_query,
                db_length=self.db_length,
                rev_comp=gene_info.rev_comp,
            )

            # Apply filtering criteria
            alignment_length = alignment.t_end - alignment.t_start

            if self.e_value_threshold is not None:
                if alignment.e_value > self.e_value_threshold:
                    continue

            if self.alignment_length_threshold is not None:
                if alignment_length < self.alignment_length_threshold:
                    continue

            # Convert to AlignmentEntryNT
            alignment_entry = AlignmentEntryNT(
                target_id=alignment.target_id,
                alignment_score=alignment.alignment_score,
                seq_identity=alignment.seq_identity,
                e_value=alignment.e_value,
                q_start=alignment.q_start,
                q_end=alignment.q_end,
                q_len=alignment.q_len,
                t_start=alignment.t_start,
                t_end=alignment.t_end,
                t_len=alignment.t_len,
                cigar=alignment.cigar,
                rev_comp=alignment.rev_comp,
                species=target_gene.species,
                locus=target_gene.locus,
                q_seq=current_query,
                t_seq=target_gene.sequence,
                reading_frame=target_gene.reading_frame if isinstance(target_gene, Gene) else None,
            )

            alignments.append(alignment_entry)

        return alignments

    # Compatibility method to match GeneAligner interface
    def _align_sequences(
        self, query: str, prefiltering_result: SpeciesPrefilteringSegmentResult
    ) -> Sequence[AlignmentEntryNT]:
        """Align sequences method compatible with GeneAligner interface"""
        all_alignments: list[AlignmentEntryNT] = []

        # Align each segment separately and collect all alignments
        for segment in prefiltering_result.segments:
            segment_alignments = self._align_segment_genes(segment, query, prefiltering_result.rev_comp_query)
            all_alignments.extend(segment_alignments)

        return all_alignments

    def align(self, query: str, both_strains: bool = True) -> Sequence[AlignmentSegment]:
        """
        Perform segment-based alignment of the query sequence.

        Args:
            query: Query sequence to align
            both_strains: Whether to consider both forward and reverse complement

        Returns:
            AlignmentResult containing alignments for each detected segment
        """
        # Get segment-centric prefiltering results
        prefiltering_result = self._prefilter(query, both_strains)

        segment_alignments = []

        # Align each segment separately
        for segment_no, segment in enumerate(prefiltering_result.segments):
            current_segment_algn_start = segment.segment_start
            # has next segment
            if segment_no + 1 < len(prefiltering_result.segments):
                next_segment_algn_start = prefiltering_result.segments[segment_no + 1].segment_start
                assert (
                    next_segment_algn_start > current_segment_algn_start
                ), f"Next segment start is before current segment start: {next_segment_algn_start} < {current_segment_algn_start}"
            else:
                next_segment_algn_start = len(query)

            alignments = self._align_segment_genes(
                segment,
                query[current_segment_algn_start:next_segment_algn_start],
                prefiltering_result.rev_comp_query[current_segment_algn_start:next_segment_algn_start],
            )

            if alignments:
                segment_alignment = AlignmentSegment(
                    segment_start=current_segment_algn_start, segment_end=next_segment_algn_start, alignments=alignments
                )
                segment_alignments.append(segment_alignment)

        return segment_alignments


class SegmentGeneAlignerAA:
    """
    Amino acid version of the segment-based gene aligner.

    Compatible with GeneAlignerAA interface for easy replacement.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        genes: Sequence[GeneAA],  # gene_id -> AA sequence mapping
        kmer_size: int = 3,  # Smaller k-mer for amino acids
        distance_threshold: int = 5,
        top_n: int = 10,
        modulo_n: int = 1,
        e_value_threshold: Optional[float] = None,
        alignment_length_threshold: Optional[int] = None,
        max_cdr3_length: Optional[int] = None,  # Added for compatibility with GeneAlignerAA
        min_prefiltering_coverage: int = 20,
        min_segment_length: int = 60,
    ) -> None:
        self.prefiltering = MultiSpeciesSegmentPrefiltering(
            genes,
            kmer_size,
            distance_threshold,
            top_n,
            modulo_n,
            min_segment_length=min_segment_length,
            min_coverage=min_prefiltering_coverage,
        )

        # Store gene sequences for alignment
        self.db_length = sum(len(gene.sequence) for gene in genes)
        self.gene_lookup = create_gene_lookup(genes)

        # Alignment parameters
        self.e_value_threshold = e_value_threshold
        self.alignment_length_threshold = alignment_length_threshold
        self.max_cdr3_length = max_cdr3_length  # Added for compatibility

    # Compatibility method to match GeneAlignerAA interface
    def _prefilter(self, query: str) -> SpeciesPrefilteringSegmentResult:
        """Prefilter method compatible with GeneAlignerAA interface"""
        return self.prefiltering.calculate_segment_matches(query)

    def _align_segment_genes_aa(self, segment: SegmentMatch, query: str) -> Sequence[AlignmentEntryAA]:
        """Align the top genes for a specific segment (amino acid version)"""
        alignments = []

        # Create aligner for this query
        aligner = StripedSmithWaterman(query, **AA_ALIGNER_PARAMS)

        for gene_info in segment.matching_genes:
            # Note: AA version doesn't support reverse complement
            if gene_info.rev_comp:
                continue

            # Get target sequence
            target_sequence = self.gene_lookup[gene_info.species_gene_id].sequence

            # Perform alignment
            alignment = align_aa(
                aligner=aligner,
                target_id=gene_info.gene_id,
                target=target_sequence,
                query=query,
                db_length=self.db_length,
            )

            # Apply filtering criteria
            alignment_length = alignment.t_end - alignment.t_start

            if self.e_value_threshold is not None:
                if alignment.e_value and alignment.e_value > self.e_value_threshold:
                    continue

            if self.alignment_length_threshold is not None:
                if alignment_length < self.alignment_length_threshold:
                    continue

            # Apply max_cdr3_length filter for compatibility with GeneAlignerAA
            if self.max_cdr3_length is not None:
                if alignment.q_start > self.max_cdr3_length:
                    continue

            # Convert to AlignmentEntryAA
            alignment_entry = AlignmentEntryAA(
                target_id=alignment.target_id,
                alignment_score=alignment.alignment_score,
                seq_identity=alignment.seq_identity,
                e_value=alignment.e_value,
                q_start=alignment.q_start,
                q_end=alignment.q_end,
                t_start=alignment.t_start,
                t_end=alignment.t_end,
                cigar=alignment.cigar,
                species=gene_info.species,
                locus=gene_info.locus,
                q_seq=query,
                t_seq=target_sequence,
            )

            alignments.append(alignment_entry)

        return alignments

    # Compatibility method to match GeneAlignerAA interface
    def _align_sequences(
        self, query: str, prefiltering_result: SpeciesPrefilteringSegmentResult
    ) -> Sequence[AlignmentEntryAA]:
        """Align sequences method compatible with GeneAlignerAA interface"""
        all_alignments: list[AlignmentEntryAA] = []

        # Align each segment separately and collect all alignments
        for segment in prefiltering_result.segments:
            segment_alignments = self._align_segment_genes_aa(segment, query)
            all_alignments.extend(segment_alignments)

        return all_alignments

    def align(self, query: str) -> Sequence[AlignmentSegment]:
        """
        Perform segment-based amino acid alignment of the query sequence.

        Args:
            query: Query AA sequence to align

        Returns:
            List[AlignmentSegment] containing alignments for each detected segment
        """
        # Get segment-centric prefiltering results (AA doesn't use rev comp)
        prefiltering_result = self._prefilter(query)

        segment_alignments = []
        # Align each segment separately
        for segment_no, segment in enumerate(prefiltering_result.segments):
            current_segment_algn_start = segment.segment_start
            if segment_no + 1 < len(prefiltering_result.segments):
                next_segment_algn_start = prefiltering_result.segments[segment_no + 1].segment_start
            else:
                next_segment_algn_start = len(query)

            alignments = self._align_segment_genes_aa(
                segment, query[current_segment_algn_start:next_segment_algn_start]
            )

            segment_alignment = AlignmentSegment(
                segment_start=current_segment_algn_start, segment_end=next_segment_algn_start, alignments=alignments
            )
            segment_alignments.append(segment_alignment)

        return segment_alignments


def get_aligner_params(germline_gene: GermlineGene, locus: Optional[Locus]) -> dict:
    """Get aligner parameters for different germline gene types"""
    # ideally this should be read from a file or env

    match germline_gene:
        case GermlineGene.V:
            return {
                "kmer_size": 9,
                "distance_threshold": 13,
                "top_n": int(os.environ.get("TOP_N", 12)),
                "modulo_n": 2,
                "e_value_threshold": 0.001,
                "alignment_length_threshold": 80,
                "min_prefiltering_coverage": 60,
                "min_segment_length": 120,
            }
        case _:
            return gene_aligner_params(germline_gene, locus)


def get_aa_aligner_params(germline_gene: GermlineGene) -> dict:
    """Get amino acid aligner parameters for different germline gene types"""
    # ideally this should be read from a file or env

    match germline_gene:
        case GermlineGene.V:
            return {
                "kmer_size": 3,
                "distance_threshold": 4,
                "top_n": int(os.environ.get("TOP_N", 100)),
                "modulo_n": 1,
                "e_value_threshold": 1e-55,
                "alignment_length_threshold": 80,
                "min_prefiltering_coverage": 20,
                "min_segment_length": 60,
            }
        case _:
            return gene_aa_aligner_params(germline_gene)


def create_v_gene_aligner(allowed_species: Sequence[Organism], db_dir: Path = GENE_DB_DIR) -> SegmentGeneAligner:
    """Create V gene segment aligner"""
    genes = []

    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS, Organism.VICUGNA_PACOS]

    gene_parsing_function = get_gene_parsing_function(GermlineGene.V)

    for species in allowed_species:
        input_path = Path(db_dir) / "gene_db" / "v_genes" / f"{species.value}.fasta"
        species_genes = read_genes(input_path, gene_parsing_function)
        genes.extend(species_genes)

    aligner_params = get_aligner_params(GermlineGene.V, None)
    genes_aligner = SegmentGeneAligner(genes, **aligner_params)
    return genes_aligner


def create_aa_v_gene_aligner(allowed_species: Sequence[Organism], aa_genes_dir: Path) -> SegmentGeneAlignerAA:
    """Create amino acid V gene segment aligner"""
    genes = []

    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS, Organism.VICUGNA_PACOS]

    for species in allowed_species:
        input_path = aa_genes_dir / "v_genes" / f"{species.value}.fasta"

        assert input_path.exists(), f"Input germline file: {str(input_path)} does not exists"

        species_genes = read_genes(input_path, partial(parse_gene_aa, species=species))
        genes.extend(species_genes)

    aligner_params = get_aa_aligner_params(GermlineGene.V)
    genes_aligner = SegmentGeneAlignerAA(genes, **aligner_params)
    return genes_aligner



import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Optional, Sequence

from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.alignment_utils import infer_reading_frame, translate
from riot_na.alignment.gene_aligner import (
    GeneAlignerAA,
    create_aa_c_gene_aligner,
    create_aa_j_gene_aligner,
    create_aa_v_gene_aligner,
)
from riot_na.alignment.segment_gene_aligner import (
    AA_ALIGNER_PARAMS,
    SegmentGeneAlignerAA,
)
from riot_na.alignment.segment_gene_aligner import (
    create_aa_v_gene_aligner as create_segment_aa_v_gene_aligner,
)
from riot_na.alignment.skbio_alignment import align_aa
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AlignmentEntryAA,
    AlignmentsAA,
    AlignmentSegment,
    AlignmentsNT,
    InternalAlignmentEntryAA,
    Locus,
    Organism,
    TranslatedAlignmentsNT,
)

logger = logging.getLogger(__name__)


class VJCAlignerAA:
    def __init__(
        self,
        v_aligner: SegmentGeneAlignerAA | GeneAlignerAA,
        j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]],
        c_aligners: Optional[dict[Organism, dict[Locus, GeneAlignerAA]]] = None,
    ) -> None:
        self.v_aligner = v_aligner
        self.j_aligners = j_aligners
        self.c_aligners = c_aligners

    def produce_aa_alignments(self, query: str) -> list[AlignmentsAA]:
        results = []

        v_segment_alignments: Sequence[AlignmentSegment] = self.v_aligner.align(query)

        for v_aa_segment_alignment in v_segment_alignments:  # pylint: disable=too-many-nested-blocks
            try:
                result = self._produce_vjc_aa_alignment(v_aa_segment_alignment)
                results.append(result)
            except ValueError as e:
                logger.error(
                    "Error aligning segment %s due to %s",
                    (
                        v_aa_segment_alignment.best_alignment.q_seq
                        if v_aa_segment_alignment.best_alignment
                        else "unknown sequence"
                    ),
                    e,
                )
        return results

    def _produce_vjc_aa_alignment(self, v_aa_segment_alignment: AlignmentSegment) -> AlignmentsAA:
        result = AlignmentsAA()
        result.segment_start = v_aa_segment_alignment.segment_start
        result.segment_end = v_aa_segment_alignment.segment_end

        v_aa_alignment = v_aa_segment_alignment.best_alignment
        if v_aa_alignment is None:
            return result

        result.v = v_aa_alignment

        query_j = v_aa_alignment.q_seq[v_aa_alignment.q_end :]
        if query_j:
            # align j genes
            species = v_aa_alignment.species
            locus = v_aa_alignment.locus

            j_aligner = self.j_aligners[species][locus]
            j_aa_alignment = j_aligner.align(query_j)[0].best_alignment

            if j_aa_alignment is None:
                return result

            result.j = j_aa_alignment

            if self.c_aligners:
                query_c = query_j[j_aa_alignment.q_end :]
                if query_c:
                    c_aligner = self.c_aligners[species][locus]
                    assert c_aligner is not None
                    c_aa_alignment = c_aligner.align(query_c)[0].best_alignment
                    if c_aa_alignment is not None:
                        result.c = c_aa_alignment

        return result


class VJCAlignmentTranslatorAA:
    def __init__(
        self,
        v_aligner: GeneAlignerAA,
        j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]],
        c_aligners: Optional[dict[Organism, dict[Locus, GeneAlignerAA]]] = None,
    ) -> None:
        self.v_aligner = v_aligner
        self.j_aligners = j_aligners
        self.c_aligners = c_aligners

    def translate_nt_alignments(  # pylint: disable=too-many-statements
        self, alignments: AlignmentsNT, extend_alignment: bool = True
    ) -> Optional[TranslatedAlignmentsNT]:
        if alignments.v:
            # Query might have been reverse complemented during V alignment.
            query_sequence = alignments.v.q_seq
            v_alignment = alignments.v
            assert v_alignment.reading_frame is not None
            reading_frame = infer_reading_frame(t_start=v_alignment.t_start, t_frame=v_alignment.reading_frame)

            # translate sequence to aa
            alignment_start = alignments.v.q_start
            alignment_end = alignments.v.q_end

            if alignments.c:
                alignment_end = alignments.c.q_end
            elif alignments.j:
                alignment_end = alignments.j.q_end

            aligned_sequence = query_sequence[alignment_start:alignment_end]
            aligned_sequence_aa = translate(aligned_sequence, reading_frame)

            aligned_sequence_no_c = query_sequence[
                alignment_start : alignments.j.q_end if alignments.j else alignments.v.q_end
            ]
            alignes_sequence_no_c_aa = translate(aligned_sequence_no_c, reading_frame)

            species = v_alignment.species
            locus = v_alignment.locus
            # Note: SegmentGeneAlignerAA doesn't have gene_lookup attribute like GeneAlignerAA
            # We'll need to access the gene sequences directly from the aligner
            v_gene = self.v_aligner.gene_lookup.get(v_alignment.lookup_gene_id())
            v_gene_aa = v_gene.sequence if v_gene else ""

            locus = v_alignment.locus

            j_gene_id = alignments.j.target_id if alignments.j else None
            j_gene_aa = None
            if j_gene_id:
                assert alignments.j
                oragnism_locus_j_aligner = self.j_aligners[species][locus]
                assert oragnism_locus_j_aligner is not None
                j_gene = oragnism_locus_j_aligner.gene_lookup.get(alignments.j.lookup_gene_id())
                j_gene_aa = j_gene.sequence if j_gene else ""

            v_aligner = StripedSmithWaterman(aligned_sequence_aa, **AA_ALIGNER_PARAMS)
            v_aa_alignment = align_aa(
                v_aligner,
                aligned_sequence_aa,
                alignments.v.target_id,
                v_gene_aa,
                calculate_score=False,
                extend_n_terminus=extend_alignment,
            )
            assert (
                v_aa_alignment.t_start != -1
                and v_aa_alignment.t_end != -1
                and v_aa_alignment.q_start != -1
                and v_aa_alignment.q_end != -1
            ), "Query translation failed: could not align sequence_aa to germline_aa"

            masked_aa = alignes_sequence_no_c_aa[v_aa_alignment.q_end :]
            j_aligner = StripedSmithWaterman(masked_aa, **AA_ALIGNER_PARAMS)
            j_aa_alignment = (
                align_aa(
                    j_aligner,
                    masked_aa,
                    j_gene_id,
                    j_gene_aa,
                    calculate_score=False,
                    extend_c_terminus=extend_alignment,
                )
                if j_gene_id and j_gene_aa
                else None
            )

            if j_aa_alignment is not None and (
                j_aa_alignment.q_start == -1
                or j_aa_alignment.q_end == -1
                or j_aa_alignment.t_start == -1
                or j_aa_alignment.t_end == -1
            ):
                j_aa_alignment = None

            extended_v_aa_alignment = AlignmentEntryAA(
                target_id=v_alignment.target_id,
                alignment_score=v_aa_alignment.alignment_score,
                seq_identity=v_aa_alignment.seq_identity,
                e_value=v_aa_alignment.e_value,
                q_start=v_aa_alignment.q_start,
                q_end=v_aa_alignment.q_end,
                t_start=v_aa_alignment.t_start,
                t_end=v_aa_alignment.t_end,
                cigar=v_aa_alignment.cigar,
                species=v_alignment.species,
                locus=v_alignment.locus,
                q_seq=aligned_sequence_aa,
                t_seq=v_gene_aa,
            )

            if j_aa_alignment:
                assert j_gene_aa
                assert j_gene_id

                extended_j_aa_alignment = AlignmentEntryAA(
                    target_id=j_gene_id,
                    alignment_score=j_aa_alignment.alignment_score,
                    seq_identity=j_aa_alignment.seq_identity,
                    e_value=j_aa_alignment.e_value,
                    q_start=j_aa_alignment.q_start,
                    q_end=j_aa_alignment.q_end,
                    t_start=j_aa_alignment.t_start,
                    t_end=j_aa_alignment.t_end,
                    cigar=j_aa_alignment.cigar,
                    species=v_alignment.species,
                    locus=v_alignment.locus,
                    q_seq=masked_aa,
                    t_seq=j_gene_aa,
                )
            else:
                extended_j_aa_alignment = None

            masked_aa_for_c, c_gene_id, c_gene_aa, c_aa_alignment = self.generate_c_aa_alignment(
                alignments, aligned_sequence_aa, species, locus, extended_j_aa_alignment, v_aa_alignment.q_end
            )

            if c_aa_alignment:
                assert c_gene_aa
                assert c_gene_id

                extended_c_aa_alignment = AlignmentEntryAA(
                    target_id=c_gene_id,
                    alignment_score=c_aa_alignment.alignment_score,
                    seq_identity=c_aa_alignment.seq_identity,
                    e_value=c_aa_alignment.e_value,
                    q_start=c_aa_alignment.q_start,
                    q_end=c_aa_alignment.q_end,
                    t_start=c_aa_alignment.t_start,
                    t_end=c_aa_alignment.t_end,
                    cigar=c_aa_alignment.cigar,
                    species=v_alignment.species,
                    locus=v_alignment.locus,
                    q_seq=masked_aa_for_c,
                    t_seq=c_gene_aa,
                )
            else:
                extended_c_aa_alignment = None

            aa_alignments = AlignmentsAA(
                v=extended_v_aa_alignment, j=extended_j_aa_alignment, c=extended_c_aa_alignment
            )

            return TranslatedAlignmentsNT(
                translated_query=aligned_sequence_aa, reading_frame=reading_frame, aa_alignments=aa_alignments
            )

        return None

    def generate_j_aa_alignment(
        self,
        aligned_sequence_aa: str,
        j_gene_id: Optional[str],
        j_gene_aa: Optional[str],
        v_aa_alignment: InternalAlignmentEntryAA,
    ) -> tuple[str, InternalAlignmentEntryAA | None]:
        masked_aa = aligned_sequence_aa[v_aa_alignment.q_end :]
        j_aligner = StripedSmithWaterman(masked_aa, **AA_ALIGNER_PARAMS)
        j_aa_alignment = (
            align_aa(j_aligner, masked_aa, j_gene_id, j_gene_aa, calculate_score=False)
            if j_gene_id and j_gene_aa
            else None
        )

        if j_aa_alignment is not None and (
            j_aa_alignment.q_start == -1
            or j_aa_alignment.q_end == -1
            or j_aa_alignment.t_start == -1
            or j_aa_alignment.t_end == -1
        ):
            j_aa_alignment = None
        return masked_aa, j_aa_alignment

    def generate_c_aa_alignment(
        self,
        alignments: AlignmentsNT,
        aligned_sequence_aa: str,
        species: Organism,
        locus: Locus,
        extended_j_aa_alignment: AlignmentEntryAA | None,
        j_offset: int,
    ) -> tuple[str, str | None, str | None, InternalAlignmentEntryAA | None]:
        if alignments.c and extended_j_aa_alignment:
            masked_aa_for_c = aligned_sequence_aa[extended_j_aa_alignment.q_end + j_offset :]

            c_gene_id = alignments.c.target_id

            assert self.c_aligners
            c_gene = self.c_aligners[species][locus].gene_lookup[alignments.c.lookup_gene_id()]
            c_gene_aa = c_gene.sequence

            c_aligner = StripedSmithWaterman(masked_aa_for_c, **AA_ALIGNER_PARAMS)
            c_aa_alignment = align_aa(
                c_aligner, masked_aa_for_c, alignments.c.target_id, c_gene_aa, calculate_score=False
            )
        else:
            masked_aa_for_c = ""
            c_gene_id = None
            c_gene_aa = None
            c_aa_alignment = None
        return masked_aa_for_c, c_gene_id, c_gene_aa, c_aa_alignment


def create_vjc_aligner_aa(
    allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR, use_segment_aligner: bool = False
) -> VJCAlignerAA:
    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS, Organism.VICUGNA_PACOS]

    aa_genes_dir = db_dir / "gene_db" / "aa_genes_deduplicated"
    v_aligner = (
        create_segment_aa_v_gene_aligner(allowed_species=allowed_species, aa_genes_dir=aa_genes_dir)
        if use_segment_aligner
        else create_aa_v_gene_aligner(allowed_species=allowed_species, aa_genes_dir=aa_genes_dir)
    )

    j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        for locus in Locus:
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            j_aligner = create_aa_j_gene_aligner(organism=organism, locus=locus, aa_genes_dir=aa_genes_dir)
            organism_aligners = j_aligners.get(organism, {})
            organism_aligners[locus] = j_aligner
            j_aligners[organism] = organism_aligners

    c_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        if organism not in [Organism.HOMO_SAPIENS, Organism.VICUGNA_PACOS, Organism.CUSTOM]:
            c_organism = Organism.HOMO_SAPIENS
        else:
            c_organism = organism

        c_organism_aligner = c_aligners.get(organism, {})

        for locus in Locus:

            # at the moment we only have heavy genes
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            c_aligner = create_aa_c_gene_aligner(organism=c_organism, locus=locus, aa_genes_dir=aa_genes_dir)
            c_organism_aligner[locus] = c_aligner
        c_aligners[organism] = c_organism_aligner

    vjc_aligner = VJCAlignerAA(v_aligner, j_aligners, c_aligners)

    return vjc_aligner


def create_vjc_alignment_translator_aa(
    allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR
) -> VJCAlignmentTranslatorAA:
    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS, Organism.VICUGNA_PACOS]

    aa_genes_dir = db_dir / "gene_db" / "aa_genes"
    v_aligner = create_aa_v_gene_aligner(allowed_species=allowed_species, aa_genes_dir=aa_genes_dir)

    j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        for locus in Locus:
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            j_aligner = create_aa_j_gene_aligner(organism=organism, locus=locus, aa_genes_dir=aa_genes_dir)
            organism_aligners = j_aligners.get(organism, {})
            organism_aligners[locus] = j_aligner
            j_aligners[organism] = organism_aligners

    c_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        c_organism_aligner = c_aligners.get(organism, {})

        if organism not in [Organism.HOMO_SAPIENS, Organism.VICUGNA_PACOS, Organism.CUSTOM]:
            c_organism = Organism.HOMO_SAPIENS
        else:
            c_organism = organism

        for locus in Locus:

            # at the moment we only have heavy genes
            if organism == Organism.VICUGNA_PACOS and locus != Locus.IGH:
                continue

            c_aligner = create_aa_c_gene_aligner(organism=c_organism, locus=locus, aa_genes_dir=aa_genes_dir)
            c_organism_aligner[locus] = c_aligner
        c_aligners[organism] = c_organism_aligner

    vj_aligner = VJCAlignmentTranslatorAA(v_aligner, j_aligners, c_aligners)

    return vj_aligner


if __name__ == "__main__":
    # case 1 detect organism

    # given query
    SAMPLE_QUERY = "QVQLQQWGAGLLKPSETLSLTCAVFGGSFSGYYWSWIRQPPGKGLEWIGEINHRGNTNDNPSLKSRVTISVDTSKNQFALKLSSVTAADTAVYYCARERGYTYGNFDHWGQGTLVTVSSASTKGPSVFPLAPSSKSTSGGTAALGCLVKDYFPEPVTVSWNSGALTSGVHTFPAVLQSSGLYSLSSVVTVPSSSLGTQTYICNVNHKPSNTKVDKKVEPKSCDKTHTCPPCPAPELLGGPSVFLFPPKPKDTLMISRTPEVTCVVVDVSHEDPEVKFNWYVDGVEVHNAKTKPREEQYNSTYRVVSVLTVLHQDWLNGKEYKCKVSNKALPAPIEKTISKAKGQPREPQVYTLPPSRDELTKNQVSLTCLVKGFYPSDIAVEWESNGQPENNYKTTPPVLDSDGSFFLYSKLTVDKSRWQQGNVFSCSVMHEALHNHYTQKSLSLSPGK"

    vjc_alnr = create_vjc_aligner_aa(allowed_species=[Organism.HOMO_SAPIENS])

    local_aa_alignments = vjc_alnr.produce_aa_alignments(SAMPLE_QUERY)
    print(json.dumps([asdict(aa_alignment) for aa_alignment in local_aa_alignments], indent=4))



import itertools
import multiprocessing as mp
from pathlib import Path
from typing import Any, Optional

from Bio import SeqIO
from psutil import cpu_count
from tqdm import tqdm

from riot_na.api.riot_numbering import create_riot_aa, create_riot_nt
from riot_na.common.io import count_fasta_records, write_airr_iter_to_csv
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    InputType,
    Organism,
    Scheme,
    SegmentedAirrRearrangementEntryAA,
    SegmentedAirrRearrangementEntryNT,
)


class _WorkerNT:
    def __init__(
        self,
        allowed_species: Optional[list[Organism]] = None,
        scheme: Scheme = Scheme.IMGT,
        db_dir: Path = GENE_DB_DIR,
        return_all_domains: bool = False,
    ) -> None:
        self.numbering = create_riot_nt(
            allowed_species=allowed_species, db_dir=db_dir, return_all_domains=return_all_domains
        )
        self.scheme = scheme

    def __call__(self, fasta_record: SeqIO.SeqRecord) -> AirrRearrangementEntryNT | list[AirrRearrangementEntryNT]:
        try:
            res = self.numbering.run_on_sequence(
                fasta_record.description,
                str(fasta_record.seq),
                self.scheme,
            )
        except Exception as exc:
            print(fasta_record.description)
            print(self.scheme)
            print(f"seq: {fasta_record.seq}")
            raise exc

        return res


class _WorkerAA:
    def __init__(
        self,
        allowed_species: Optional[list[Organism]] = None,
        scheme: Scheme = Scheme.IMGT,
        db_dir: Path = GENE_DB_DIR,
        return_all_domains: bool = False,
        extend_alignment: bool = False,
    ) -> None:
        self.numbering = create_riot_aa(
            allowed_species=allowed_species, db_dir=db_dir, return_all_domains=return_all_domains
        )
        self.scheme = scheme
        self.extend_alignment = extend_alignment

    def __call__(self, fasta_record: SeqIO.SeqRecord) -> AirrRearrangementEntryAA | list[AirrRearrangementEntryAA]:
        try:
            res = self.numbering.run_on_sequence(
                fasta_record.description,
                str(fasta_record.seq),
                self.scheme,
                extend_alignment=self.extend_alignment,
            )
        except Exception as exc:
            print(fasta_record.description)
            print(self.scheme)
            print(f"seq: {fasta_record.seq}")
            raise exc

        return res


def _worker_initializer(
    allowed_species: Optional[list[Organism]] = None,
    scheme: Scheme = Scheme.IMGT,
    db_dir: Path = GENE_DB_DIR,
    input_type: InputType = InputType.NT,
    return_all_domains: bool = False,
    extend_alignment: bool = False,
):
    global worker  # pylint: disable=global-statement,global-variable-undefined
    if input_type == InputType.NT:
        worker = _WorkerNT(allowed_species, scheme, db_dir, return_all_domains)  # type: ignore
    else:
        worker = _WorkerAA(allowed_species, scheme, db_dir, return_all_domains, extend_alignment)  # type: ignore


def _worker_call(*args, **kwds):
    return worker(*args, **kwds)  # type: ignore


def run_on_file_mp(  # pylint: disable=too-many-arguments
    db_dir: Path,
    input_fasta_path: Path,
    result_path: Path,
    n_processes: int = cpu_count(logical=False),
    input_format: str = "fasta",
    scheme: Scheme = Scheme.IMGT,
    allowed_species: Optional[list[Organism]] = None,
    input_type: InputType = InputType.NT,
    limit: Optional[int] = None,
    return_all_domains: bool = False,
    extend_alignment: bool = False,
):
    with mp.Pool(
        processes=n_processes,
        initializer=_worker_initializer,
        initargs=(allowed_species, scheme, db_dir, input_type, return_all_domains, extend_alignment),
    ) as pool:
        result_iter = tqdm(
            pool.imap(_worker_call, itertools.islice(SeqIO.parse(input_fasta_path, input_format), limit)),
            total=count_fasta_records(input_fasta_path, input_format=input_format),
        )
        record_type: Any
        match input_type:
            case InputType.NT:
                record_type = AirrRearrangementEntryNT if not return_all_domains else SegmentedAirrRearrangementEntryNT
            case InputType.AA:
                record_type = AirrRearrangementEntryAA if not return_all_domains else SegmentedAirrRearrangementEntryAA

        write_airr_iter_to_csv(result_path, record_type, result_iter)


if __name__ == "__main__":
    # Example usage

    # .write.text("/home/pawel.dudzic/workspace/analyzer/projects/automation/experiments/exploration/new_therapeutics/paper/new_mess/data/13_08_2025/molecules_segments.fasta", lineSep="\n")
    INPUT_PATH = "/home/pawel.dudzic/workspace/analyzer/projects/automation/experiments/exploration/new_therapeutics/paper/new_mess/data/13_08_2025/molecules_segments.fasta"
    OUTPUT_PATH = "/home/pawel.dudzic/workspace/analyzer/projects/automation/experiments/exploration/new_therapeutics/paper/new_mess/data/13_08_2025/molecules_segments_numbered.csv"
    run_on_file_mp(
        db_dir=GENE_DB_DIR,
        input_fasta_path=Path(INPUT_PATH),
        result_path=Path(OUTPUT_PATH),
        n_processes=8,
        input_format="fasta",
        scheme=Scheme.IMGT,
        allowed_species=[Organism.HOMO_SAPIENS],
        input_type=InputType.AA,
    )



# pylint: disable=too-many-branches, too-many-statements

from pathlib import Path
from typing import Optional

from cachetools import cached

from riot_na.airr.airr_builder import AirrBuilder, SegmentedAirrBuilder
from riot_na.airr.airr_builder_aa import AirrBuilderAA, SegmentedAirrBuilderAA
from riot_na.alignment.aa_gene_alignments import (
    VJCAlignerAA,
    VJCAlignmentTranslatorAA,
    create_vjc_aligner_aa,
    create_vjc_alignment_translator_aa,
)
from riot_na.alignment.nt_gene_alignments import VDJCAlignerNT, create_vdjc_aligner_nt
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    ChainType,
    Organism,
    Scheme,
)
from riot_na.schemes.region_offsets import (
    infer_aa_region_offsets,
    infer_nt_region_offsets,
    infer_region_offsets,
)
from riot_na.schemes.scheme_alignment import (
    SchemeAligner,
    fix_imgt_cdrs_numbering,
    get_positional_scheme_mapping,
    get_scheme_residue_mapping,
    produce_numbering,
    sort_imgt_numbering,
)


class RiotNumberingNT:
    def __init__(
        self,
        vdjc_aligner_nt: VDJCAlignerNT,
        vjc_alignment_translator_aa: VJCAlignmentTranslatorAA,
        scheme_aligner: SchemeAligner,
        return_all_domains: bool = False,
    ):
        self.vdjc_aligner = vdjc_aligner_nt
        self.vjc_alignment_translator_aa = vjc_alignment_translator_aa
        self.scheme_aligner = scheme_aligner
        self.return_all_domains = return_all_domains

    def run_on_sequence(
        self,
        header: str,
        query_sequence: str,
        scheme: Scheme = Scheme.IMGT,
        extend_alignment: bool = True,
    ) -> AirrRearrangementEntryNT | list[AirrRearrangementEntryNT]:  # pylint: disable=too-many-statements
        """
        Align and number nucleotide sequence.

        Parameters:
        -----------
        - header : Fasta sequence header, usually this is simply sequence ID.
        - query_sequence : Nucleotide sequence to number.
        - scheme : Which numbering scheme to use (default IMGT).
        - return_all_domains: If True, return all domains of multiple domain proteins.
        - extend_alignment: Extend alignments to cover full sequence.

        Returns:
        -----------
        AirrRearrangementEntryNT object with alignment results (Extended AIRR format)
        """
        query_sequence = query_sequence.upper()
        sequence = ""
        segment_start = -1
        segment_end = -1
        translated_alignments = None
        nt_offsets = None
        aa_offsets = None
        sch_alignment = None
        positional_scheme_mapping = None
        numbering = None
        scheme_alignment_exc = None

        # produce alignments for v,d,j,c
        alignments = self.vdjc_aligner.produce_nt_alignments(query_sequence)

        # Query might have been reverse complemented during V alignment.
        results = []

        for alignment in alignments:
            if alignment.v:
                sequence = alignment.v.q_seq
                segment_start = alignment.segment_start or 0
                segment_end = alignment.segment_end or len(sequence)
                try:
                    translated_alignments = self.vjc_alignment_translator_aa.translate_nt_alignments(
                        alignment, extend_alignment=extend_alignment
                    )

                    if translated_alignments:
                        sch_alignment = self.scheme_aligner.align_to_scheme(
                            translated_alignments.aa_alignments, scheme, extend_alignment=False
                        )

                        # calculate aa offsets based on the numbering (iterate over cigar to produce positional-> imgt mapping)
                        assert sch_alignment
                        assert alignment.v
                        chain_type = ChainType.from_locus(alignment.v.locus)

                        region_offsets = infer_region_offsets(sch_alignment, scheme, chain_type)
                        aa_offsets = infer_aa_region_offsets(region_offsets)

                        alignment_start = alignment.v.q_start

                        nt_offsets = infer_nt_region_offsets(
                            region_offsets, alignment_start, translated_alignments.reading_frame
                        )
                        numbering = produce_numbering(
                            translated_alignments.translated_query, sch_alignment.alignment_str, sch_alignment.q_start
                        )

                        if scheme == Scheme.IMGT:
                            numbering = fix_imgt_cdrs_numbering(numbering)
                            numbering = sort_imgt_numbering(numbering)

                        scheme_residue_mapping = get_scheme_residue_mapping(numbering)
                        positional_scheme_mapping = get_positional_scheme_mapping(
                            scheme_residue_mapping, sch_alignment.q_start
                        )
                except AssertionError as err:
                    scheme_alignment_exc = str(err)

            # produce airr result
            # this is separated on purpose to enable other output formats
            airr_builder = AirrBuilder(header, sequence or query_sequence, scheme)
            if self.return_all_domains:
                airr_builder = SegmentedAirrBuilder(header, sequence, scheme, query_sequence)
                airr_builder.with_segment_start_end(segment_start, segment_end)

            if alignment.v:
                airr_builder.with_v_gene_alignment(alignment.v)

            if translated_alignments:
                airr_builder.with_aa_alignments(translated_alignments)

            if alignment.j:
                airr_builder.with_j_gene_alignment(alignment.j)

            if alignment.d:
                airr_builder.with_d_gene_alignment(alignment.d)

            if alignment.c:
                airr_builder.with_c_gene_alignment(alignment.c)

            if sch_alignment:
                assert nt_offsets
                assert aa_offsets
                airr_builder.with_nt_region_offsets(nt_offsets)
                airr_builder.with_aa_region_offsets(aa_offsets)
                airr_builder.with_aa_scheme_alignment(sch_alignment)

            if numbering:
                (
                    airr_builder.with_scheme_residue_mapping(scheme_residue_mapping).with_positional_scheme_mapping(
                        positional_scheme_mapping
                    )
                )

            if scheme_alignment_exc:
                airr_builder.with_scheme_alignment_exc(scheme_alignment_exc)

            result = airr_builder.build()
            results.append(result)

        if not results:
            results = [
                (
                    AirrBuilder(header, sequence, scheme).build()
                    if not self.return_all_domains
                    else SegmentedAirrBuilder(header, sequence, scheme, query_sequence).build()
                )
            ]

        if self.return_all_domains:
            return results

        return sorted(results)[-1]


class RiotNumberingAA:
    def __init__(self, vjc_aligner_aa: VJCAlignerAA, scheme_aligner: SchemeAligner, return_all_domains: bool = False):
        self.vjc_aligner_aa = vjc_aligner_aa
        self.scheme_aligner = scheme_aligner
        self.return_all_domains = return_all_domains

    def run_on_sequence(
        self,
        header: str,
        query_sequence: str,
        scheme: Scheme = Scheme.IMGT,
        extend_alignment: bool = True,
    ) -> AirrRearrangementEntryAA | list[AirrRearrangementEntryAA]:  # pylint: disable=too-many-statements
        """
        Align and number amino acid sequence.

        Parameters:
        -----------
        - header : Fasta sequence header, usually this is simply sequence ID.
        - query_sequence : Amino acid sequence to number.
        - scheme : Which numbering scheme to use (default IMGT).
        - extend_alignment: RIOT uses Striped Smith-Waterman local alignemtn algorithm.
        Due to that, if query sequence has too many differences (due to mutations or other reasons)
        to the germline sequence at 5' or / and 3' ends the result may not show that part of the gene.
        This option appends missing amino acids, so the result covers full gene sequence. Beware that
        those query fragments are not truly a part of alignemnt and are shown only for convenience!
        - return_all_domains: If True, return all domains of multiple domain proteins.

        Returns:
        -----------
        AirrRearrangementEntryAA object with alignment results (Extended AIRR format)
        """
        query_sequence = query_sequence.upper()
        sequence_aa = query_sequence
        sch_alignment = None
        aa_offsets = None
        positional_scheme_mapping = None
        numbering = None
        scheme_alignment_exc = None
        segment_start = -1
        segment_end = -1

        # produce alignments for v,j,c
        alignments = self.vjc_aligner_aa.produce_aa_alignments(query_sequence)

        results = []

        for alignment in alignments:
            if alignment.v:
                segment_start = alignment.segment_start or 0
                segment_end = alignment.segment_end or len(alignment.v.q_seq)

                sequence_aa = alignment.v.q_seq
                try:
                    sch_alignment = self.scheme_aligner.align_to_scheme(
                        alignment, scheme, extend_alignment=extend_alignment
                    )

                    # calculate aa offsets based on the numbering (iterate over cigar to produce positional-> imgt mapping)
                    assert sch_alignment
                    chain_type = ChainType.from_locus(alignment.v.locus)

                    region_offsets = infer_region_offsets(sch_alignment, scheme, chain_type)
                    aa_offsets = infer_aa_region_offsets(region_offsets)

                    numbering = produce_numbering(alignment.v.q_seq, sch_alignment.alignment_str, sch_alignment.q_start)

                    if scheme == Scheme.IMGT:
                        numbering = fix_imgt_cdrs_numbering(numbering)
                        numbering = sort_imgt_numbering(numbering)

                    scheme_residue_mapping = get_scheme_residue_mapping(numbering)
                    positional_scheme_mapping = get_positional_scheme_mapping(
                        scheme_residue_mapping, sch_alignment.q_start
                    )
                except AssertionError as err:
                    scheme_alignment_exc = str(err)

            # produce airr result
            # this is separated on purpose to enable other output formats
            airr_builder = AirrBuilderAA(header, query_sequence, scheme)
            if self.return_all_domains:
                airr_builder = SegmentedAirrBuilderAA(header, sequence_aa, scheme, query_sequence)
                airr_builder.with_segment_start_end(segment_start, segment_end)

            if alignment.v:
                airr_builder.with_v_gene_alignment_aa(alignment.v)

            if alignment.j:
                airr_builder.with_j_gene_alignment_aa(alignment.j)

            if alignment.c:
                airr_builder.with_c_gene_alignment_aa(alignment.c)

            if sch_alignment:
                assert aa_offsets
                airr_builder.with_aa_region_offsets(aa_offsets)
                airr_builder.with_aa_scheme_alignment(sch_alignment)

            if numbering:
                (
                    airr_builder.with_scheme_residue_mapping(scheme_residue_mapping).with_positional_scheme_mapping(
                        positional_scheme_mapping
                    )
                )

            if scheme_alignment_exc:
                airr_builder.with_scheme_alignment_exc(scheme_alignment_exc)

            result = airr_builder.build()
            results.append(result)

        if not results:
            results = [
                (
                    AirrBuilderAA(header, sequence_aa, scheme).build()
                    if not self.return_all_domains
                    else SegmentedAirrBuilderAA(header, sequence_aa, scheme, query_sequence).build()
                )
            ]

        if self.return_all_domains:
            return results

        return sorted(results)[-1]


def create_riot_nt(
    allowed_species: Optional[list[Organism]] = None,
    return_all_domains: bool = False,
    db_dir: Path = GENE_DB_DIR,
) -> RiotNumberingNT:
    """
    Factory function for creating RiotNumberingNT object.

    Parameters:
    -----------
    - allowed_species : Limit gene database to specified Organism list (default use all species - mouse and human).
    - return_all_domains: If True, return all domains of multiple domain proteins.
    - db_dir : Path to gene and scheme mappings database directory (default use embeded database).

    Returns:
    -----------
    RiotNumberingNT object used for performing alignment and numbering.
    """
    vdjc_aligner_nt = create_vdjc_aligner_nt(
        allowed_species=allowed_species, db_dir=db_dir, use_segment_aligner=return_all_domains
    )
    vjc_alignment_translator_aa = create_vjc_alignment_translator_aa(allowed_species=allowed_species, db_dir=db_dir)
    scheme_aligner = SchemeAligner(allowed_species=allowed_species, db_dir=db_dir)
    return RiotNumberingNT(vdjc_aligner_nt, vjc_alignment_translator_aa, scheme_aligner, return_all_domains)


def create_riot_aa(
    allowed_species: Optional[list[Organism]] = None,
    return_all_domains: bool = False,
    db_dir: Path = GENE_DB_DIR,
) -> RiotNumberingAA:
    """
    Factory function for creating RiotNumberingAA object.

    Parameters:
    -----------
    - allowed_species : Limit gene database to specified Organism list (default use all species - mouse and human).
    - return_all_domains: If True, return all domains of multiple domain proteins.
    - db_dir : Path to gene and scheme mappings database directory (default use embeded database).

    Returns:
    -----------
    RiotNumberingAA object used for performing alignment and numbering.
    """
    vjc_aligner_aa = create_vjc_aligner_aa(
        allowed_species=allowed_species,
        db_dir=db_dir,
        use_segment_aligner=return_all_domains,
    )
    scheme_aligner = SchemeAligner(allowed_species=allowed_species, db_dir=db_dir)
    return RiotNumberingAA(vjc_aligner_aa, scheme_aligner, return_all_domains)


@cached({})
def get_or_create_riot_nt(**riot_kwargs) -> RiotNumberingNT:
    return create_riot_nt(**riot_kwargs)


@cached({})
def get_or_create_riot_aa(**riot_kwargs) -> RiotNumberingAA:
    return create_riot_aa(**riot_kwargs)


if __name__ == "__main__":
    import json
    from dataclasses import asdict

    # vdj_alnr = create_vdjc_aligner_nt()
    # vjc_aa_translator = create_vjc_alignment_translator_aa()
    scheme_alnr = SchemeAligner()

    # riot_numbering = RiotNumberingNT(vdj_alnr, vjc_aa_translator, scheme_alnr)
    # # given query
    # QUERY = "CAGGTGCAGCTCGTGGAGTCTGGGGGAGGCTTGGTGCAGCCTGGGGGGTCTCTGAGACTCTCCTGTGTAGCTTCTGGATTCACCTTCAGTGACTATGCCATGACCTGGCACCGCCAGGCCTCAGGGTCGGAGCTCGAGTTGGTCGCGGCGATTAGCAATGGCCGTGGTCGTAACACAAATTATGCAGATGCCGTGAAGGGTCGATTCACCATCTCCAGAGACAATGCCAAGAACACGGTGTATCTGCAAATGAACAACCTGAAATCTGACGACACGGCCGTGTACTACTGTAACGCGGCCGGGTTGGGGCCCGAATATGACTACTGGGGCCAGGGGACCCAGGTCACCGTCTCCGCGGAACCCAAGACACCAAAACCACAA"
    # sample_result = riot_numbering.run_on_sequence("header", QUERY, Scheme.IMGT)
    # print(json.dumps(asdict(sample_result), indent=4))

    vjc_aa_alnr = create_vjc_aligner_aa()
    aa_numbering = RiotNumberingAA(vjc_aa_alnr, scheme_alnr)

    AA_QUERY = "QVQLVESGGGLVQPGGSLRLSCVASGFTFSDYAMTWHRQASGSELELVAAISNGRGRNTNYADAVKGRFTISRDNAKNTVYLQMNNLKSDDTAVYYCNAAGLGPEYDYWGQGTQVTVSAEPKTPKPQ"

    aa_sample_result = aa_numbering.run_on_sequence("header", AA_QUERY, Scheme.IMGT)
    # Handle both single result and list result
    if isinstance(aa_sample_result, list):
        print(json.dumps([asdict(result) for result in aa_sample_result], indent=4))
    else:
        print(json.dumps(asdict(aa_sample_result), indent=4))



"""
Utility functions for handling gene matches and species information.

This module provides shared functionality for converting between different gene match formats
and extracting species information, avoiding the need for proxy classes and code duplication.
"""

from typing import Sequence

from riot_na.data.model import Gene, GeneAA


def create_gene_lookup(genes: Sequence[Gene | GeneAA]) -> dict[str, Gene | GeneAA]:
    """Create a lookup dictionary for genes using species|name as key."""
    return {f"{gene.species.value}|{gene.locus.value}|{gene.name}": gene for gene in genes}



from typing import Dict

import pytest

import riot_na


class TestPrefilteringNucleotideAPI:
    """Test suite for the Prefiltering API with nucleotide sequences."""

    @pytest.fixture
    def basic_nucleotide_genes(self) -> Dict[str, str]:
        """Basic nucleotide gene database for testing."""
        return {
            "IGHV1-69*01": "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA",
            "IGHV3-23*01": "GAAGTGCAGCTGCTGGAGTCTGGCGGAGGATTGGTACAGCCTGGCGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTTAGCAGTTACTGGATGAGCTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTCTCAGCTATTAATGGTGGTAGCGGCACTTATTACGCTGACTCTGTGAAGGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTTTATCTGCAAATGAACAGCCTGAGAGCTGAGGACACGGCTGTGTATTATTGTGCTAAGGGTGAT",
            "IGHV4-34*01": "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGGAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGTTACACCTTTACCAGTTACGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGCTGGATCAACGCTGGCAACGGTAACACAAAATATTCACAGAAGTTCCAGGGCAGAGTCACCATTACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGAT",
            "IGKV1-27*01": "GACATCCAGATGACCCAGTCTCCATCCTCCCTGTCTGCATCTGTAGGAGACAGAGTCACCATCACTTGCCGGGCAAGTCAGATCAGTAGCAGCTATTTAAATTGGTATCAGCAGAAACCAGGGAAAGCCCCTAAGCTCCTGATCTATGCTGCATCCAGTTTGCAAAGTGGGGTCCCATCAAGGTTCAGTGGCAGTGGATCTGGGACAGATTTCACTCTCACCATCAGCAGTCTGCAACCTGAAGATTTTGCAACTTATTACTGTCAGCAATCTTATAGTACTCCTCTCACTTTCGGCGGAGGGACCAAGGTGGAGATCAAACGG",
            "IGLV2-8*01": "CAGTCTGTGTTGACGCAGCCGCCCTCAGTGTCTGGGGCCCCAGGGCAGAGGGTCACCATCTCCTGCACTGGGAGCAGCTCCAACATCGGGGCAGGTTATGATGTACACTGGTACCAGCAGCTTCCAGGAACAGCCCCCAAACTCCTCATCTATGGTAACAGCAATCGGCCCTCAGGGGTCCCTGACCGATTCTCTGGCTCCAAGTCTGGCACCTCAGCCTCCCTGGCCATCAGTGGGCTCCAGTCTGAGGATGAGGCTGATTATTACTGTGCAACATGGGATGACAGCCTGAATGGTCCG",
        }

    @pytest.fixture
    def nucleotide_immunoglobulin_genes(self) -> Dict[str, str]:
        """Comprehensive nucleotide immunoglobulin gene database."""
        return {
            # Heavy chain V genes (nucleotide)
            "IGHV1-2*01": "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA",
            "IGHV3-33*01": "GAGGTGCAGCTGGTGGAGTCTGGGGGAGGCTTGGTCCAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCAGTAGCTACGCTATGCACTGGGTCCGCCAGGCTCCAGGCAAGGGGCTGGAGTGGGTGGCAGTTATATCATATGATGGAAGTAATAAATACTATGCAGACTCCGTGAAGGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTGTATTACTGTGCGAGAGGGGGGTATA",
            "IGHV4-39*01": "CAGGTGCAGCTGCAGGAGTCGGGCCCAGGACTGGTGAAGCCTTCGGAGACCCTGTCCCTCACCTGCGCTGTCTATGGTGGGTCCTTCAGTGGTTACTACTGGAGCTGGATCCGCCAGCCCCCAGGGAAGGGGCTGGAGTGGATTGGGGAAATCAATCATAGTGGAAGCACCAACTACAACCCGTCCCTCAAGAGTCGAGTCACCATATCAGTAGACACGTCCAAGAACCAGTTCTCCCTGAAGCTGAGCTCTGTGACCGCCGCGGACACGGCCGTGTATTACTGTGCGAGAGATCGGGGGAGGAA",
            # Kappa light chain V genes (nucleotide)
            "IGKV1-39*01": "GACATCCAGATGACCCAGTCTCCATCCTCCCTGTCTGCATCTGTAGGAGACAGAGTCACCATCACTTGCCGGGCAAGTCAGGGCATTAGTAGCTGGCTGGCCTGGTATCAGCAGAAACCAGGGAAAGCCCCTAAGCTCCTGATCTATGCTGCATCCAGTTTGCAAAGTGGGGTCCCATCAAGGTTCAGTGGCAGTGGATCTGGGACAGATTTCACTCTCACCATCAGCAGTCTGCAACCTGAAGATTTTGCAACTTACTACTGTCAGCAAGCTAACAGCTTCCCTTATACGTTCGGCCAAGGGACCAAGGTG",
            "IGKV2-28*01": "GATGTTGTGATGACTCAGTCTCCACTCTCCCTGCCCGTCACCCTTGGACAGCCGGCCTCCATCTCCTGCAGGTCTAGTCAAAGCCTCGTAGACAGTGATGGAAACACCTATTTGAATTGGTTGCTACAGAGGCCAGGCCAGTCTCCAAAGCGCCTAATCTATCTGGTGTCTAAACTGGACTCTGGAGTCCCTGACAGGTTCACTGGCAGTGGATCAGGGACAGATTTCACACTGAAAATCAGCAGAGTGGAGGCTGAGGATGTGGGTGTTTATTACTGCATGCAAGCTCTACAAACTCCGTACACTTTT",
            "IGKV3-15*01": "GAAATTGTGTTGACACAGTCTCCAGCCACCCTGTCTTTGTCTCCAGGGGAAAGAGCCACCCTCTCCTGCAGGGCCAGTCAGAGTGTTAGCAGCAGCTACTTAGCCTGGTACCAGCAGAAACCTGGCCAGGCTCCCAGGCTCCTCATCTATGGTGCATCCAGCAGGGCCACTGGCATCCCAGACAGGTTCAGTGGCAGTGGGTCTGGGACAGACTTCACTCTCACCATCAGCAGCCTAGAGCCTGAAGATTTTGCAGTTTATTACTGTCAGCAGTATGGTAGCTCACCTCCGTGGACGTTCGGCCAAGGG",
            # Lambda light chain V genes (nucleotide)
            "IGLV1-44*01": "CAGTCTGCCCTGACTCAGCCTGCCTCCGTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGACGTTGGTGGTTATAACTATGTCTCCTGGTACCAACAACACCCAGGCAAAGCCCCCAAACTCATGATTTATGACGTCAGTAAGCGGCCCTCAGGGGTTTCTAATCGCTTCTCTGGCTCCAAGTCTGGCAACACGGCCTCCCTGACCATCTCTGGGCTCCAGGCTGAGGATGAGGCTGATTATTACTGCAGCTCATATACAAGCAGCAGCACTCTGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCTA",
            "IGLV2-14*01": "CAGTCTGTGCTGACTCAGCCACCCTCAGCGTCTGGGACCCCCGGGCAGAGGGTCACCATCTCTTGTTCTGGAAGCAGCTCCAACATCGGAAGTAATACTGTAAACTGGTACCAGCAGCTCCCAGGAACGGCCCCCAAACTCCTCATCTATAGTAATAATCAGCGGCCCTCAGGGGTCCCTGACCGATTCTCTGGCTCCAAGTCTGGCACGTCAGCCACCCTGGGCATCACCGGACTCCAGACTGGGGACGAGGCCGATTATTACTGCGGAACATGGGATAGCAGCCTGAGTGCTGTGGTATTCGGCGGAGGGACCAAGCTGACCGTCCTA",
            # J genes (nucleotide)
            "IGHJ4*01": "ACTACTTTGACTACTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
            "IGHJ6*01": "ATTACTACTACTACTACGGTATGGACGTCTGGGGCCAAGGGACCACGGTCACCGTCTCCTCAG",
            "IGKJ1*01": "TGGACGTTCGGCCAAGGGACCAAGGTGGAAATCAAACG",
            "IGLJ3*01": "TGGACGTTCGGCCAAGGGACCAAGGTGGAAATCAAACG",
        }

    @pytest.fixture
    def multi_domain_nucleotide_sequence(self) -> str:
        """A nucleotide sequence containing multiple domains for testing segment detection."""
        return (
            # Heavy chain V domain (nucleotide)
            "GAGGTGCAGCTGGTGGAGTCTGGGGGAGGCTTGGTCCAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCAGTAGCTACGCTATGCACTGGGTCCGCCAGGCTCCAGGCAAGGGGCTGGAGTGGGTGGCAGTTATATCATATGATGGAAGTAATAAATACTATGCAGACTCCGTGAAGGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTGTATTACTGTGCGAGAGGGGGGTATA"
            # Linker (nucleotide)
            "GGAGGCGGTTCAGGCGGAGGTGGCTCTGGCGGTGGCGGATCG"
            # Kappa light chain V domain (nucleotide)
            "GAAATTGTGTTGACACAGTCTCCAGCCACCCTGTCTTTGTCTCCAGGGGAAAGAGCCACCCTCTCCTGCAGGGCCAGTCAGAGTGTTAGCAGCAGCTACTTAGCCTGGTACCAGCAGAAACCTGGCCAGGCTCCCAGGCTCCTCATCTATGGTGCATCCAGCAGGGCCACTGGCATCCCAGACAGGTTCAGTGGCAGTGGGTCTGGGACAGACTTCACTCTCACCATCAGCAGCCTAGAGCCTGAAGATTTTGCAGTTTATTACTGTCAGCAGTATGGTAGCTCACCTCCGTGGACGTTCGGCCAAGGG"
            # Another linker (nucleotide)
            "GGAGGCGGTTCAGGCGGAGGTGGCTCTGGCGGTGGCGGATCG"
            # Lambda light chain V domain (nucleotide)
            "CAGTCTGCCCTGACTCAGCCTGCCTCCGTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGACGTTGGTGGTTATAACTATGTCTCCTGGTACCAACAACACCCAGGCAAAGCCCCCAAACTCATGATTTATGACGTCAGTAAGCGGCCCTCAGGGGTTTCTAATCGCTTCTCTGGCTCCAAGTCTGGCAACACGGCCTCCCTGACCATCTCTGGGCTCCAGGCTGAGGATGAGGCTGATTATTACTGCAGCTCATATACAAGCAGCAGCACTCTGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCTA"
        )

    def test_prefiltering_instantiation_basic(self, basic_nucleotide_genes):
        """Test basic Prefiltering instantiation with nucleotide genes."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_nucleotide_genes, kmer_size=6, distance_threshold=1, top_n=5, modulo_n=1
        )
        assert prefiltering is not None

    def test_prefiltering_instantiation_with_min_segment_length(self, basic_nucleotide_genes):
        """Test Prefiltering instantiation with min_segment_length parameter."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_nucleotide_genes, kmer_size=6, distance_threshold=1, top_n=5, modulo_n=1, min_segment_length=15
        )
        assert prefiltering is not None

    def test_calculate_top_matches_basic(self, basic_nucleotide_genes):
        """Test basic top matches calculation with nucleotide sequences."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_nucleotide_genes, kmer_size=6, distance_threshold=1, top_n=3, modulo_n=1
        )

        # Use a nucleotide query that should match one of our genes
        query = "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA"

        result = prefiltering.calculate_top_matches(query)

        # Verify result structure
        assert result is not None
        assert result.query == query
        assert result.rev_comp_query == "-"  # No reverse complement in this method
        assert isinstance(result.top_matches, list)
        assert len(result.top_matches) <= 3  # Should respect top_n limit

        # Check that at least one match was found
        if len(result.top_matches) > 0:
            match = result.top_matches[0]
            assert hasattr(match, "gene_id")
            assert hasattr(match, "coverage")
            assert hasattr(match, "rev_comp")
            assert match.rev_comp is False  # Forward orientation only

    def test_calculate_top_matches_with_rev_comp(self, basic_nucleotide_genes):
        """Test top matches calculation with reverse complement."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_nucleotide_genes, kmer_size=6, distance_threshold=1, top_n=5, modulo_n=1
        )

        query = "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA"

        result = prefiltering.calculate_top_matches_with_rev_comp(query)

        # Verify result structure
        assert result is not None
        assert result.query == query
        assert result.rev_comp_query != "-"  # Should have reverse complement
        assert len(result.rev_comp_query) == len(query)  # Same length as original
        assert isinstance(result.top_matches, list)
        assert len(result.top_matches) <= 5  # Should respect top_n limit

    def test_calculate_segment_matches_basic(self, nucleotide_immunoglobulin_genes):
        """Test basic segment matches calculation with nucleotides."""
        prefiltering = riot_na.Prefiltering(
            genes=nucleotide_immunoglobulin_genes,
            kmer_size=6,
            distance_threshold=1,
            top_n=10,
            modulo_n=1,
            min_segment_length=20,
        )

        # Use a nucleotide sequence that should match multiple genes
        query = "GAGGTGCAGCTGGTGGAGTCTGGGGGAGGCTTGGTCCAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCAGTAGCTACGCTATGCACTGGGTCCGCCAGGCTCCAGGCAAGGGGCTGGAGTGGGTGGCAGTTATATCATATGATGGAAGTAATAAATACTATGCAGACTCCGTGAAGGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTGTATTACTGTGCGAGAGGGGGGTATA"

        result = prefiltering.calculate_segment_matches(query)

        # Verify result structure
        assert result is not None
        assert result.query == query
        assert result.rev_comp_query == "-"  # No reverse complement in this method
        assert isinstance(result.segments, list)

        # Check segment structure if any found
        if len(result.segments) > 0:
            segment = result.segments[0]
            assert hasattr(segment, "query_start")
            assert hasattr(segment, "query_end")
            assert hasattr(segment, "coverage")
            assert hasattr(segment, "match_count")
            assert hasattr(segment, "matching_genes")
            assert hasattr(segment, "min_target_start")

            # Check segment bounds
            assert segment.query_start <= segment.query_end
            assert segment.query_end <= len(query)
            assert isinstance(segment.matching_genes, list)

    def test_multi_domain_segment_detection(self, nucleotide_immunoglobulin_genes, multi_domain_nucleotide_sequence):
        """Test segment detection on multi-domain nucleotide sequences."""
        prefiltering = riot_na.Prefiltering(
            genes=nucleotide_immunoglobulin_genes,
            kmer_size=6,
            distance_threshold=1,
            top_n=10,
            modulo_n=1,
            min_segment_length=15,
        )

        result = prefiltering.calculate_segment_matches_with_rev_comp(multi_domain_nucleotide_sequence)

        # Should detect multiple segments
        assert len(result.segments) > 0

        # Check that segments don't overlap (should be non-overlapping)
        segments = sorted(result.segments, key=lambda s: s.query_start)
        for i in range(len(segments) - 1):
            assert segments[i].query_end <= segments[i + 1].query_start, "Segments should not overlap"

    def test_query_length_handling(self, basic_nucleotide_genes):
        """Test behavior with nucleotide queries of different lengths."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_nucleotide_genes, kmer_size=6, distance_threshold=1, top_n=5, modulo_n=1
        )

        # Test minimum valid query length (exactly kmer_size)
        min_query = "ACGTAC"  # 6 characters, exactly kmer_size
        result = prefiltering.calculate_top_matches(min_query)
        assert result is not None
        assert result.query == min_query

        # Test longer query
        long_query = "ACGTACGTACGTACGTACGTACGTACGTACGTACGT"
        result = prefiltering.calculate_top_matches(long_query)
        assert result is not None
        assert result.query == long_query


class TestPrefilteringAminoAcidAPI:
    """Test suite for the Prefiltering API with amino acid sequences."""

    @pytest.fixture
    def basic_amino_acid_genes(self) -> Dict[str, str]:
        """Basic amino acid gene database for testing."""
        return {
            "IGHV1-69*01": "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR",
            "IGHV3-23*01": "EVQLLESGGGLVQPGGSLRLSCAASGFTFSSYAMSWVRQAPGKGLEWVSAISGSGGSTYYADSVKGRFTISRDNSKNTLYLQMNSLRAEDTAVYYCAK",
            "IGHV4-34*01": "QVQLVQSGGGVKKPGASVKVSCKASGYTFTSYAMHWVRQAPGQGLEWMGWINAGNGNTKYSQKFQGRVTITRDTSASTAYMELSSLRSEDTAVYYCAR",
            "IGKV1-27*01": "DIQMTQSPSSLSASVGDRVTITCRASQSISSYLNWYQQKPGKAPKLLIYAASSLQSGVPSRFSGSGSGTDFTLTISSLQPEDFATYYCQQSYSTPLT",
            "IGLV2-8*01": "QSALTQPASVSGSPGQSITISCTGTSSDVGGYNYVSWYQQHPGKAPKLMIYDVSNRPSGVSNRFSGSKSGNTASLTISGLQAEDEADYYCSSYTSSSTLV",
        }

    @pytest.fixture
    def amino_acid_immunoglobulin_genes(self) -> Dict[str, str]:
        """Comprehensive amino acid immunoglobulin gene database."""
        return {
            # Heavy chain V genes (amino acid)
            "IGHV1-2*01": "QVQLVQSGAEVKKPGSSVKVSCKASGYTFTSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR",
            "IGHV1-69-2*01": "EVQLVQSGAEVKKPGATVKISCKVSGYTFTDYYMHWVQQAPGKGLEWMGLVDPEDGETIYAEKFQGRVTITADTSTDTAYMELSSLRSEDTAVYYCAT",
            "IGHV3-33*01": "EVQLVESGGGLVQPGGSLRLSCAASGFTFSSYAMHWVRQAPGKGLEWVAVIYHGSNKYYANPVKGRFTISRDNAKNSLYLQMNSLRAEDTAVYYCAR",
            "IGHV1-3*01": "QVQLVQSGAEVKKPGASVKVSCKASGYTFTSYAMHWVRQAPGQRLEWMGWINAGNGNTKYSQKFQGRVTITRDTSASTAYMELSSLRSEDTAVYYCAR",
            "IGHV4-39*01": "QVQLQESGPGLVKPSETLSLTCAVYGGSFSGYYWSWIRQPPGKGLEWIGEINHSGSTNYPDSLKSRVISIVDTSKNQFSLKLSSVTAADTAVYYCAR",
            # Kappa light chain V genes (amino acid)
            "IGKV1-39*01": "DIQMTQSPSSLSASVGDRVTITCRASQGISSWLAWYQQKPGKAPKLLIYAASSLQSGVPSRFSGSGSGTDFTLTISSLQPEDFATYYCQQANSFPYT",
            "IGKV2-28*01": "DVVMTQSPLSLPVTLGQPASISCRSSQSLVHSGATYLNWFQQRPGQSPRRLLIYKVSNRDSGVPDRFTGSGSGTDFTLKISRVEAEDVGVYYCAR",
            "IGKV3-15*01": "EIVLTQSPATLSLSPGERATLSCRASQSVSSSYLAWYQQKPGQAPRLLIYGASSRATGIPDRFSGSGSGTDFTLTISRLEPEDFAVYYCQQYGSSPPWT",
            # Lambda light chain V genes (amino acid)
            "IGLV1-44*01": "QSALTQPASVSGSPGQSITISCTGTSSDVGGYNYVSWYQQHPGKAPKLMIYDVSNRPSGVSNRFSGSKSGNTASLTISGLQAEDEADYYCSSYTSSSTLV",
            "IGLV2-14*01": "QSALTQPASVSGSPGQSITISCTGTSSDVGGYNYVSWYQQHPGKAPKLMIYDVSNRPSGVSNRFSGSKSGNTASLTISGLQAEDEADYYCGTWDSSLSAVVFGGGTKLT",
            # J genes (amino acid)
            "IGHJ4*01": "YYYYGMDVWGQGTLVTVSS",
            "IGHJ6*01": "YYYYYGMDVWGQGTLVTVSS",
            "IGKJ1*01": "WTFGQGTKVEIK",
            "IGLJ3*01": "WTFGQGTKVEIK",
        }

    @pytest.fixture
    def multi_domain_amino_acid_sequence(self) -> str:
        """An amino acid sequence containing multiple domains for testing segment detection."""
        return (
            # Heavy chain V domain (amino acid)
            "EVQLVESGGGLVQPGGSLRLSCAASGFTFSSYAMHWVRQAPGKGLEWVAVIYHGSNKYYANPVKGRFTISRDNAKNSLYLQMNSLRAEDTAVYYCAR"
            # Linker (amino acid)
            "GGGSGGGSGGGSGGS"
            # Kappa light chain V domain (amino acid)
            "EIVLTQSPATLSLSPGERATLSCRASQSVSSSYLAWYQQKPGQAPRLLIYGASSRATGIPDRFSGSGSGTDFTLTISRLEPEDFAVYYCQQYGSSPPWT"
            # Another linker (amino acid)
            "GGGSGGGSGGGSGGS"
            # Lambda light chain V domain (amino acid)
            "QSALTQPASVSGSPGQSITISCTGTSSDVGGYNYVSWYQQHPGKAPKLMIYDVSNRPSGVSNRFSGSKSGNTASLTISGLQAEDEADYYCSSYTSSSTLV"
        )

    def test_calculate_top_matches_basic(self, basic_amino_acid_genes):
        """Test basic top matches calculation with amino acid sequences."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_amino_acid_genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=1,
            top_n=3,
            modulo_n=1,
        )

        # Use an amino acid query that should match one of our genes
        query = "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR"

        result = prefiltering.calculate_top_matches(query)

        # Verify result structure
        assert result is not None
        assert result.query == query
        assert result.rev_comp_query == "-"  # No reverse complement in this method
        assert isinstance(result.top_matches, list)
        assert len(result.top_matches) <= 3  # Should respect top_n limit

        # Check that at least one match was found
        if len(result.top_matches) > 0:
            match = result.top_matches[0]
            assert hasattr(match, "gene_id")
            assert hasattr(match, "coverage")
            assert hasattr(match, "rev_comp")
            assert match.rev_comp is False  # Forward orientation only

    def test_calculate_segment_matches_basic(self, amino_acid_immunoglobulin_genes):
        """Test basic segment matches calculation with amino acids."""
        prefiltering = riot_na.Prefiltering(
            genes=amino_acid_immunoglobulin_genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=3,
            top_n=10,
            modulo_n=1,
            min_segment_length=60,  # Shorter for amino acids
        )

        # Use an amino acid sequence that should match multiple genes
        query = "EIQLQQSGPELGKPGASVKVSCRASGFSFADYYIYWVKQSHGKSLELIGYIDPFNGGDTYNQIFKGKATLTVDKSSSTAFMYLNSLTSEDSAVYYCAAFRNPSFDFWGQGTTLTVSSGGGGGGGGSGGGGGGGGGQIVLIQSPPIMSASPGEKVTLTCSASSSVSSRYLYWYQQKPGSSPKLWIYGTSNLASGVPARFSGSGSGTSFSLTISSMEAEDAASYFCHQWSSFPFTFGSGTKLEIK"

        result = prefiltering.calculate_segment_matches(query)

        # Verify result structure
        assert result is not None
        assert result.query == query
        assert result.rev_comp_query == "-"  # No reverse complement in this method
        assert isinstance(result.segments, list)
        assert len(result.segments) >= 1 and len(result.segments) <= 2

        # Check segment structure if any found
        if len(result.segments) > 0:
            segment = result.segments[0]
            assert hasattr(segment, "query_start")
            assert hasattr(segment, "query_end")
            assert hasattr(segment, "coverage")
            assert hasattr(segment, "match_count")
            assert hasattr(segment, "matching_genes")
            assert hasattr(segment, "min_target_start")

            # Check segment bounds
            assert segment.query_start <= segment.query_end
            assert segment.query_end <= len(query)
            assert isinstance(segment.matching_genes, list)

    def test_multi_domain_segment_detection(self, amino_acid_immunoglobulin_genes, multi_domain_amino_acid_sequence):
        """Test segment detection on multi-domain amino acid sequences."""
        prefiltering = riot_na.Prefiltering(
            genes=amino_acid_immunoglobulin_genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=1,
            top_n=10,
            modulo_n=1,
            min_segment_length=8,
        )

        result = prefiltering.calculate_segment_matches_with_rev_comp(multi_domain_amino_acid_sequence)

        # Should detect multiple segments
        assert len(result.segments) > 0

        # Check that segments don't overlap (should be non-overlapping)
        segments = sorted(result.segments, key=lambda s: s.query_start)
        for i in range(len(segments) - 1):
            assert segments[i].query_end <= segments[i + 1].query_start, "Segments should not overlap"

    def test_therapeutic_scenarios(self, amino_acid_immunoglobulin_genes):
        """Test therapeutic scenarios with amino acid sequences."""
        prefiltering = riot_na.Prefiltering(
            genes=amino_acid_immunoglobulin_genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=4,
            top_n=10,
            modulo_n=1,
            min_segment_length=60,
            min_coverage=20,
        )

        query = "EIQLQQSGPELGKPGASVKVSCRASGFSFADYYIYWVKQSHGKSLELIGYIDPFNGGDTYNQIFKGKATLTVDKSSSTAFMYLNSLTSEDSAVYYCAAFRNPSFDFWGQGTTLTVSS"
        result = prefiltering.calculate_segment_matches(query)
        assert len(result.segments) > 0
        assert result.segments[0].query_start == 5
        assert result.segments[0].query_end == 97
        assert result.segments[0].coverage == 30
        assert result.segments[0].matching_genes[0].gene_id == "IGHV1-2*01"
        assert result.segments[0].matching_genes[1].gene_id == "IGHV1-3*01"
        assert result.segments[0].matching_genes[2].gene_id == "IGHV1-69-2*01"

    def test_query_length_handling(self, basic_amino_acid_genes):
        """Test behavior with amino acid queries of different lengths."""
        prefiltering = riot_na.Prefiltering(
            genes=basic_amino_acid_genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=1,
            top_n=5,
            modulo_n=1,
        )

        # Test minimum valid query length (exactly kmer_size)
        min_query = "QVQ"  # 3 characters, exactly kmer_size
        result = prefiltering.calculate_top_matches(min_query)
        assert result is not None
        assert result.query == min_query

        # Test longer query
        long_query = (
            "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR"
        )
        result = prefiltering.calculate_top_matches(long_query)
        assert result is not None
        assert result.query == long_query

        shorter_than_kmer_size_query = "QV"
        with pytest.raises(ValueError):
            prefiltering.calculate_top_matches(shorter_than_kmer_size_query)


class TestPrefilteringEdgeCases:
    """Test edge cases and practical scenarios for both sequence types."""

    def test_small_gene_database_nucleotide(self):
        """Test behavior with small nucleotide gene database."""
        small_genes = {"SINGLE_GENE": "ACGTACGTACGTACGTACGTACGTACGTACGTACGT"}

        prefiltering = riot_na.Prefiltering(genes=small_genes, kmer_size=6, distance_threshold=1, top_n=5, modulo_n=1)

        result = prefiltering.calculate_top_matches("ACGTACGTACGTACGTACGTACGTACGTACGTACGT")
        assert result is not None
        assert len(result.top_matches) <= 1  # Can't have more matches than genes

    def test_small_gene_database_amino_acid(self):
        """Test behavior with small amino acid gene database."""
        small_genes = {
            "SINGLE_GENE": "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR"
        }

        prefiltering = riot_na.Prefiltering(
            genes=small_genes, kmer_size=3, distance_threshold=1, top_n=5, modulo_n=1  # Appropriate for amino acids
        )

        result = prefiltering.calculate_top_matches(
            "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR"
        )
        assert result is not None
        assert len(result.top_matches) <= 1  # Can't have more matches than genes

    def test_segment_vs_top_matches_consistency_nucleotide(self):
        """Test that segment matching gives same results as top_matches for single-domain nucleotide sequences."""
        genes = {
            "IGHV1-69*01": "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA",
            "IGHV3-23*01": "GAAGTGCAGCTGCTGGAGTCTGGCGGAGGATTGGTACAGCCTGGCGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTTAGCAGTTACTGGATGAGCTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTCTCAGCTATTAATGGTGGTAGCGGCACTTATTACGCTGACTCTGTGAAGGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTTTATCTGCAAATGAACAGCCTGAGAGCTGAGGACACGGCTGTGTATTATTGTGCTAAGGGTGAT",
            "IGKV1-27*01": "GACATCCAGATGACCCAGTCTCCATCCTCCCTGTCTGCATCTGTAGGAGACAGAGTCACCATCACTTGCCGGGCAAGTCAGATCAGTAGCAGCTATTTAAATTGGTATCAGCAGAAACCAGGGAAAGCCCCTAAGCTCCTGATCTATGCTGCATCCAGTTTGCAAAGTGGGGTCCCATCAAGGTTCAGTGGCAGTGGATCTGGGACAGATTTCACTCTCACCATCAGCAGTCTGCAACCTGAAGATTTTGCAACTTATTACTGTCAGCAATCTTATAGTACTCCTCTCACTTTCGGCGGAGGGACCAAGGTGGAGATCAAACGG",
        }

        prefiltering = riot_na.Prefiltering(
            genes=genes, kmer_size=6, distance_threshold=10, top_n=10, modulo_n=1, min_segment_length=20
        )

        # Test with a single-domain nucleotide sequence (should match IGHV1-69*01)
        single_domain_query = "CAGGTGCAGCTGGTGCAGTCTGGGGCTGAGGTGAAGAAGCCTGGGGCCTCAGTGAAGGTCTCCTGCAAGGCTTCTGGATACACCTTCACCAGCTATGCTATGCACTGGGTGCGACAGGCCCCTGGACAAGGGCTTGAGTGGATGGGATGGATCAACCCTAACAGTGGTGGCACAAACTATGCACAGAAGTTTCAGGGCAGGGTCACCATGACCAGGGACACGTCCATCAGCACAGCCTACATGGAGCTGAGCAGGCTGAGATCTGACGACACGGCCGTGTATTACTGTGCGAGAGA"

        # Get results from both methods
        top_matches_result = prefiltering.calculate_top_matches(single_domain_query)
        segment_result = prefiltering.calculate_segment_matches(single_domain_query)

        # Both should find results
        assert len(top_matches_result.top_matches) > 0, "Top matches should find at least one result"
        assert len(segment_result.segments) > 0, "Segment matching should find at least one segment"

        # For single domain, there should be exactly one segment that spans most/all of the query
        assert len(segment_result.segments) == 1, "Single domain should result in exactly one segment"
        segment = segment_result.segments[0]

        # The segment should span most of the query (allowing for some boundary variation)
        query_length = len(single_domain_query)
        segment_length = segment.query_end - segment.query_start
        assert (
            segment_length >= query_length * 0.8
        ), f"Segment should cover at least 80% of query length. Got {segment_length}/{query_length}"

        # The top gene matches should be consistent between methods
        # Extract gene IDs from top matches
        top_gene_ids = {match.gene_id for match in top_matches_result.top_matches}
        segment_gene_ids = {match.gene_id for match in segment.matching_genes}

        # There should be overlap in the top gene matches
        common_genes = top_gene_ids.intersection(segment_gene_ids)
        assert (
            len(common_genes) > 0
        ), f"Should have common genes between methods. Top: {top_gene_ids}, Segment: {segment_gene_ids}"

        # The best match should be the same (or very similar coverage)
        if top_matches_result.top_matches and segment.matching_genes:
            best_top_match = top_matches_result.top_matches[0]
            best_segment_match = max(segment.matching_genes, key=lambda x: x.coverage)

            # Best matches should have similar coverage (within reasonable tolerance)
            coverage_diff = abs(best_top_match.coverage - best_segment_match.coverage)
            max_coverage = max(best_top_match.coverage, best_segment_match.coverage)
            if max_coverage > 0:
                relative_diff = coverage_diff / max_coverage
                assert (
                    relative_diff <= 0.2
                ), f"Coverage should be similar. Top: {best_top_match.coverage}, Segment: {best_segment_match.coverage}"

    def test_segment_vs_top_matches_consistency_amino_acid(self):
        """Test that segment matching gives same results as top_matches for single-domain amino acid sequences."""
        genes = {
            "IGHV1-69*01": "QVQLVQSGAEVKKPGSSVKVSCKASGGTFSSYAISWVRQAPGQGLEWMGRIIPILGIANYAQKFQGRVTITADKSTSTAYMELSSLRSEDTAVYYCAR",
            "IGHV3-23*01": "EVQLLESGGGLVQPGGSLRLSCAASGFTFSSYAMSWVRQAPGKGLEWVSAISGSGGSTYYADSVKGRFTISRDNSKNTLYLQMNSLRAEDTAVYYCAK",
            "IGKV1-27*01": "DIQMTQSPSSLSASVGDRVTITCRASQSISSYLNWYQQKPGKAPKLLIYAASSLQSGVPSRFSGSGSGTDFTLTISSLQPEDFATYYCQQSYSTPLT",
            "IGHV1-2*02": "QVQLVQSGAEVKKPGASVKVSCKASGYTFTGYYMHWVRQAPGQGLEWMGWINPNSGGTNYAQKFQGRVTMTRDTSISTAYMELSRLRSDDTAVYYCAR",
            "IGKV1D-33*01": "DIQMTQSPSSLSASVGDRVTITCQASQDISNYLNWYQQKPGKAPKLLIYDASNLETGVPSRFSGSGSGTDFTFTISSLQPEDIATYYCQQYDNLP",
        }

        prefiltering = riot_na.Prefiltering(
            genes=genes,
            kmer_size=3,  # Appropriate for amino acids
            distance_threshold=4,
            top_n=10,
            modulo_n=1,
            min_segment_length=50,  # Shorter for amino acids
        )

        # Test with a single-domain amino acid sequence (should match IGHV1-2*02)
        single_domain_query = (
            "DIQMTQSPSSLSASVGDTVTITCQANGYLNWYQQRRGKAPKLLIYDGSKLERGVPSRFSGRRWGQEYNLTINNLQPEDIATYFCQVYEFVVPGTRLDLK"
        )

        # Get results from both methods
        top_matches_result = prefiltering.calculate_top_matches(single_domain_query)
        segment_result = prefiltering.calculate_segment_matches(single_domain_query)

        # Both should find results
        assert len(top_matches_result.top_matches) > 0, "Top matches should find at least one result"
        assert len(segment_result.segments) > 0, "Segment matching should find at least one segment"

        # For single domain, there should be exactly one segment that spans most/all of the query
        assert len(segment_result.segments) == 1, "Single domain should result in exactly one segment"
        segment = segment_result.segments[0]

        # The segment should span most of the query (allowing for some boundary variation)
        query_length = len(single_domain_query)
        segment_length = segment.query_end - segment.query_start
        assert (
            segment_length >= query_length * 0.8
        ), f"Segment should cover at least 80% of query length. Got {segment_length}/{query_length}"

        # The top gene matches should be consistent between methods
        # Extract gene IDs from top matches
        top_gene_ids = {match.gene_id for match in top_matches_result.top_matches[0:2]}
        segment_gene_ids = {match.gene_id for match in segment.matching_genes[0:2]}
        assert (
            top_gene_ids == segment_gene_ids
        ), f"Top gene ids should be the same. Top: {top_gene_ids}, Segment: {segment_gene_ids}"

        # There should be overlap in the top gene matches
        common_genes = top_gene_ids.intersection(segment_gene_ids)
        assert (
            len(common_genes) > 0
        ), f"Should have common genes between methods. Top: {top_gene_ids}, Segment: {segment_gene_ids}"

        # The best match should be the same (or very similar coverage)
        if top_matches_result.top_matches and segment.matching_genes:
            best_top_match = top_matches_result.top_matches[0]
            best_segment_match = max(segment.matching_genes, key=lambda x: x.coverage)

            assert (
                best_top_match.gene_id == best_segment_match.gene_id
            ), f"Best match should be the same. Top: {best_top_match.gene_id}, Segment: {best_segment_match.gene_id}"
            assert (
                best_top_match.coverage == best_segment_match.coverage
            ), f"Best match should have the same coverage. Top: {best_top_match.coverage}, Segment: {best_segment_match.coverage}"



class GeneMatch:
    gene_id: str
    rev_comp: bool
    coverage: int

class GeneSegment:
    start_target: int
    end_target: int
    start_query: int
    end_query: int
    coverage: int
    match_count: int

class SegmentMatch:
    query_start: int
    query_end: int
    coverage: int
    match_count: int
    matching_genes: list[GeneMatch]
    segment_start: int

    def query_length(self) -> int: ...

class PrefilteringResult:
    query: str
    rev_comp_query: str
    top_matches: list[GeneMatch]

class PrefilteringSegmentResult:
    query: str
    rev_comp_query: str
    segments: list[SegmentMatch]

    def domain_count(self) -> int: ...

class Prefiltering:
    def __init__(
        self,
        genes: dict[str, str],
        kmer_size: int,
        distance_threshold: int,
        top_n: int,
        modulo_n: int,
        min_segment_length: int = 30,
        min_coverage: int = 20,
    ): ...
    def calculate_top_matches_with_rev_comp(self, query: str) -> PrefilteringResult: ...
    def calculate_top_matches(self, query: str) -> PrefilteringResult: ...
    def calculate_segment_matches(self, query: str) -> PrefilteringSegmentResult: ...
    def calculate_segment_matches_with_rev_comp(self, query: str) -> PrefilteringSegmentResult: ...
